# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Product, superclass of all EOReader satellites products """
# pylint: disable=W0107
from __future__ import annotations

import datetime as dt
import functools
import gc
import logging
import os
import platform
import shutil
import tempfile
from abc import abstractmethod
from enum import unique
from io import BytesIO
from typing import Tuple, Union

import geopandas as gpd
import numpy as np
import rasterio
import validators
import xarray as xr
from affine import Affine
from lxml import etree, html
from rasterio import shutil as rio_shutil
from rasterio import transform, warp
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from sertit import AnyPath, files, logs, misc, path, rasters, strings, xml
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME, cache, utils
from eoreader.bands import (
    DEM,
    GREEN1,
    HILLSHADE,
    NEEDED_BANDS,
    SLOPE,
    BandNames,
    compute_index,
    indices,
    is_clouds,
    is_dem,
    is_index,
    is_sar_band,
    is_spectral_band,
    to_band,
    to_str,
)
from eoreader.env_vars import CI_EOREADER_BAND_FOLDER, DEM_PATH
from eoreader.exceptions import (
    InvalidBandError,
    InvalidIndexError,
    InvalidProductError,
    InvalidTypeError,
)
from eoreader.keywords import DEM_KW, HILLSHADE_KW, SLOPE_KW
from eoreader.reader import Constellation, Reader
from eoreader.stac import StacItem
from eoreader.utils import UINT16_NODATA, simplify

LOGGER = logging.getLogger(EOREADER_NAME)
PRODUCT_FACTORY = Reader()


@unique
class SensorType(ListEnum):
    """
    Sensor type of the products, optical or SAR
    """

    OPTICAL = "Optical"
    """For optical data"""

    SAR = "SAR"
    """For SAR data"""


@unique
class OrbitDirection(ListEnum):
    """
    Orbit Direction
    """

    ASCENDING = "ASCENDING"
    """Ascending sensing orbit direction"""

    DESCENDING = "DESCENDING"
    """Descending sensing orbit direction"""

    UNKNOWN = "UNKNOWN"
    """Unknown orbit direction"""


class Product:
    """Super class of EOReader Products"""

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self.needs_extraction = True
        """Does this product needs to be extracted to be processed ? (:code:`True` by default)."""

        self.path = AnyPath(product_path)
        """Usable path to the product, either extracted or archived path, according to the satellite."""

        self.filename = path.get_filename(self.path)
        """Product filename"""

        self._use_filename = False
        self.name = None
        """Product true name (as specified in the metadata)"""

        self.split_name = None
        """
        Split name, to retrieve every information from its true name (dates, tile, product type...).
        """

        self.archive_path = AnyPath(archive_path) if archive_path else self.path
        """Archive path, same as the product path if not specified.
        Useful when you want to know where both the extracted and archived version of your product are stored."""

        self.is_archived = self.path.is_file()
        """ Is the archived product is processed
        (a products is considered as archived if its products path is a directory)."""

        # The output will be given later
        self._tmp_output = None
        self._output = None
        self._remove_tmp_process = remove_tmp

        # Get the products date and datetime
        self.date = None
        """Acquisition date."""

        self.datetime = None
        """Acquisition datetime."""

        self.tile_name = None
        """Tile if possible (for data that can be piled, for example S2 and Landsats)."""

        self.sensor_type = None
        """Sensor type, SAR or optical."""

        self.product_type = None
        """Product type, satellite-related field, such as L1C or L2A for Sentinel-2 data."""

        self.instrument = None
        """Product instrument, such as MSI for Sentinel-2 data."""

        self.bands = None
        """
        Band mapping between band wrapping names such as
        :code:`GREEN` and band real number such as :code:`03` for Sentinel-2.
        """

        self.is_reference = False
        """If the product is a reference, used for algorithms that need pre and post data, such as fire detection."""

        self.corresponding_ref = []
        """The corresponding reference products to the current one
         (if the product is not a reference but has a reference data corresponding to it).
         A list because of multiple ref in case of non-stackable products (S3, S1...)"""

        self.nodata = -9999
        """ Product nodata, set to -9999 by default """

        # Mask values
        self._mask_true = 1
        self._mask_false = 0
        self._mask_nodata = 255

        self.constellation = kwargs.get("constellation")
        """Product constellation, such as Sentinel-2"""

        # Set the resolution, needs to be done when knowing the product type
        self.resolution = None
        """
        Default resolution in meters of the current product.
        For SAR product, we use Ground Range resolution as we will automatically orthorectify the tiles.
        """

        self.pixel_size = None
        """
        For SAR data, it is important to distinguish (square) pixel spacing from actual resolution.
        (see `this <https://natural-resources.canada.ca/maps-tools-and-publications/satellite-imagery-and-air-photos/tutorial-fundamentals-remote-sensing/satellites-and-sensors/spatial-resolution-pixel-size-and-scale/9407>` for more information).
        For optical data, those two terms have usually the same meaning (for a fully zoomed raster).
        """

        self.condensed_name = None
        """
        Condensed name, the filename with only useful data to keep the name unique
        (ie. :code:`20191215T110441_S2_30TXP_L2A_122756`).
        Used to shorten names and paths.
        """

        self.constellation_id = None
        """Constellation ID, i.e. :code:`S2` for :code:`Sentinel-2`"""

        self.is_ortho = True
        """True if the images are orthorectified and the footprint is retrieved easily."""

        self._stac = None

        # Manage output
        if output_path:
            self._tmp_output = None
            self._output = AnyPath(output_path)
        else:
            self._tmp_output = tempfile.TemporaryDirectory()
            self._output = AnyPath(self._tmp_output.name)

        # Pre initialization
        self._pre_init(**kwargs)

        # Only compute data if OK (for now OK is extracted if needed)
        if self.is_archived and self.needs_extraction:
            LOGGER.warning(f"{self.filename} needs to be extracted to be used !")
        else:
            # Get the product real name
            self.name = self._get_name()
            self.split_name = self._get_split_name()

            # Get the products date and datetime
            self.datetime = self.get_datetime(as_datetime=True)
            self.date = self.get_date(as_date=True)

            # Constellation and satellite ID
            if not self.constellation:
                self.constellation = self._get_constellation()
            self.constellation_id = (
                self.constellation
                if isinstance(self.constellation, str)
                else self.constellation.name
            )
            self._set_instrument()

            # Post initialization
            self._post_init(**kwargs)

            # Set product type, needs to be done after the post-initialization
            self._set_product_type()

            # Set the pixel size, needs to be done when knowing the product type
            self._set_pixel_size()

            self._map_bands()

            # Condensed name
            self.condensed_name = self._get_condensed_name()

            # Temporary file path (private)
            self._tmp_process = self._output.joinpath(f"tmp_{self.condensed_name}")
            os.makedirs(self._tmp_process, exist_ok=True)

    def __del__(self):
        """Cleaning up _tmp directory"""
        self.clear()

        # -- Remove temp folders
        if self._tmp_output:
            self._tmp_output.cleanup()

        elif self._remove_tmp_process:
            files.remove(self._tmp_process)

    @abstractmethod
    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        raise NotImplementedError

    @abstractmethod
    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        raise NotImplementedError

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get UTM footprint of the products (without nodata, *in french == emprise utile*)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
               index                                           geometry
            0      0  POLYGON ((199980.000 4500000.000, 199980.000 4...

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        raise NotImplementedError

    @cache
    @abstractmethod
    def extent(self) -> gpd.GeoDataFrame:
        """
        Get UTM extent of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.utm_extent()
                                                        geometry
            0  POLYGON ((309780.000 4390200.000, 309780.000 4...

        Returns:
            gpd.GeoDataFrame: Extent in UTM
        """
        raise NotImplementedError

    @cache
    @abstractmethod
    def crs(self) -> CRS:
        """
        Get UTM projection of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.utm_crs()
            CRS.from_epsg(32630)

        Returns:
            crs.CRS: CRS object
        """
        raise NotImplementedError

    @abstractmethod
    def _map_bands(self):
        """
        Map bands
        """
        raise NotImplementedError

    def _get_band_folder(self, writable: bool = False) -> AnyPathType:
        """
        Manage the case of CI SNAP Bands

        Returns:
            AnyPathType: Band folder
        """
        band_folder = self._tmp_process

        # Manage CI bands (when we do not write anything, read only)
        if not writable:
            ci_band_folder = os.environ.get(CI_EOREADER_BAND_FOLDER)
            if ci_band_folder:
                ci_band_folder = AnyPath(ci_band_folder)
                if ci_band_folder.is_dir():
                    # If we need a writable directory, check it
                    band_folder = ci_band_folder

        return band_folder

    @abstractmethod
    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        raise NotImplementedError

    @abstractmethod
    def _set_product_type(self) -> None:
        """
        Set product type
        """
        raise NotImplementedError

    @abstractmethod
    def _set_instrument(self) -> None:
        """
        Set product type
        """
        raise NotImplementedError

    @classmethod
    def _get_constellation(cls) -> Constellation:
        class_module = cls.__module__.split(".")[-1]
        constellation_id = class_module.replace("_product", "").upper()
        return getattr(Constellation, constellation_id)

    def _get_name(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        if (
            self._use_filename
            and self.constellation
            and Reader().valid_name(self.path, self.constellation)
        ):
            name = self.filename
        else:
            name = self._get_name_constellation_specific()

        return name

    @abstractmethod
    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        raise NotImplementedError

    @abstractmethod
    def _get_condensed_name(self) -> str:
        """
        Set product condensed name.

        Returns:
            str: Condensed name
        """
        raise NotImplementedError

    def _get_split_name(self) -> list:
        """
        Get split name (erasing empty strings in it by precaution, especially for S1 and S3 data)

        Returns:
            list: Split products name
        """
        return utils.get_split_name(self.name)

    @abstractmethod
    def get_datetime(self, as_datetime: bool = False) -> Union[str, dt.datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 8, 24, 11, 6, 31)
            >>> prod.get_datetime(as_datetime=False)
            '20200824T110631'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        raise NotImplementedError

    def get_date(self, as_date: bool = False) -> Union[str, dt.date]:
        """
        Get the product's acquisition date.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_date(as_date=True)
            datetime.datetime(2020, 8, 24, 0, 0)
            >>> prod.get_date(as_date=False)
            '20200824'

        Args:
            as_date (bool): Return the date as a datetime.date. If false, returns a string.

        Returns:
            str: Its acquisition date
        """
        date = self.get_datetime().split("T")[0]

        if as_date:
            date = strings.str_to_date(date, date_format="%Y%m%d")

        return date

    def _construct_band_path(
        self,
        band: BandNames,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        writable: bool = False,
        **kwargs,
    ) -> AnyPathType:
        """
        Get cloud band path.

        Args:
            band (BandNames): Wanted band
            pixel_size (float): Band pixel size in meters
            writable (bool): True if we want the band folder to be writeable
            kwargs: Additional arguments

        Returns:
            AnyPathType: Clean band path
        """
        # Manage pixel size
        if pixel_size is None:
            if size is not None:
                pixel_size = self._pixel_size_from_img_size(size)
            else:
                pixel_size = self.pixel_size

        # Convert to str
        res_str = self._pixel_size_to_str(pixel_size)

        return self._get_band_folder(writable).joinpath(
            f"{self.condensed_name}_{to_str(band)[0]}_{res_str.replace('.', '-')}.tif",
        )

    @abstractmethod
    def get_default_band_path(self, **kwargs) -> AnyPathType:
        """
        Get default band path (among the existing ones).

        Usually :code:`GREEN` band for optical data and the first existing one between :code:`VV` and :code:`HH` for SAR data.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_default_band_path()
            'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B03.jp2'

        Args:
            kwargs: Additional arguments

        Returns:
            AnyPathType: Default band path
        """
        raise NotImplementedError

    @abstractmethod
    def get_default_band(self) -> BandNames:
        """
        Get default band:
        Usually :code:`GREEN` band for optical data and the first existing one between :code:`VV` and :code:`HH` for SAR data.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_default_band()
            <SpectralBandNames.GREEN: 'GREEN'>


        Returns:
            str: Default band
        """
        raise NotImplementedError

    @abstractmethod
    def get_existing_bands(self) -> list:
        """
        Return the existing bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_existing_bands()
            [<SpectralBandNames.CA: 'COASTAL_AEROSOL'>,
            <SpectralBandNames.BLUE: 'BLUE'>,
            <SpectralBandNames.GREEN: 'GREEN'>,
            <SpectralBandNames.RED: 'RED'>,
            <SpectralBandNames.VRE_1: 'VEGETATION_RED_EDGE_1'>,
            <SpectralBandNames.VRE_2: 'VEGETATION_RED_EDGE_2'>,
            <SpectralBandNames.VRE_3: 'VEGETATION_RED_EDGE_3'>,
            <SpectralBandNames.NIR: 'NIR'>,
            <SpectralBandNames.NNIR: 'NARROW_NIR'>,
            <SpectralBandNames.WV: 'WATER_VAPOUR'>,
            <SpectralBandNames.CIRRUS: 'CIRRUS'>,
            <SpectralBandNames.SWIR_1: 'SWIR_1'>,
            <SpectralBandNames.SWIR_2: 'SWIR_2'>]

        Returns:
            list: List of existing bands in the products
        """
        raise NotImplementedError

    @abstractmethod
    def get_existing_band_paths(self) -> dict:
        """
        Return the existing band paths (orthorectified if needed).

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_existing_band_paths()
            {
                <SpectralBandNames.CA: 'COASTAL_AEROSOL'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B01.jp2',
                ...,
                <SpectralBandNames.SWIR_2: 'SWIR_2'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B12.jp2'
            }

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        raise NotImplementedError

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the raw band paths.

        Args:
            kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        return self.get_existing_band_paths()

    @abstractmethod
    def get_band_paths(
        self, band_list: list, pixel_size: float = None, **kwargs
    ) -> dict:
        """
        Return the paths of required bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_band_paths([GREEN, RED])
            {
                <SpectralBandNames.GREEN: 'GREEN'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B03.jp2',
                <SpectralBandNames.RED: 'RED'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B04.jp2'
            }

        Args:
            band_list (list): List of the wanted bands
            pixel_size (float): Band pixel size (in meters)
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        raise NotImplementedError

    @abstractmethod
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        Returns:
            (etree._Element, dict): Metadata XML root and its namespace
        """
        raise NotImplementedError

    def _read_mtd_xml(
        self, mtd_from_path: str, mtd_archived: str = None
    ) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dicts as a dict

        Args:
            mtd_from_path (str): Metadata regex (glob style) to find from extracted product
            mtd_archived (str): Metadata regex (re style) to find from archived product

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces

        """
        try:
            if self.is_archived:
                root = xml.read_archive(self.path, f".*{mtd_archived}")
            else:
                try:
                    try:
                        mtd_file = next(self.path.glob(f"**/*{mtd_from_path}"))
                    except ValueError:
                        mtd_file = next(self.path.glob(f"*{mtd_from_path}"))

                    try:
                        root = xml.read(mtd_file)
                    except ValueError as ex:
                        raise InvalidProductError from ex
                except StopIteration as ex:
                    raise InvalidProductError(
                        f"Metadata file ({mtd_from_path}) not found in {self.path}"
                    ) from ex
        except etree.XMLSyntaxError:
            raise InvalidProductError(f"Invalid metadata XML for {self.path}!")

        # Get namespaces map (only useful ones)
        nsmap = {key: f"{{{ns}}}" for key, ns in root.nsmap.items()}
        pop_list = ["xsi", "xs", "xlink"]
        for ns in pop_list:
            if ns in nsmap.keys():
                nsmap.pop(ns)

        return root, nsmap

    def _read_mtd_html(
        self, mtd_from_path: str, mtd_archived: str = None
    ) -> html.HtmlElement:
        """
        Read metadata and outputs the metadata HTML root

        Args:
            mtd_from_path (str): Metadata regex (glob style) to find from extracted product
            mtd_archived (str): Metadata regex (re style) to find from archived product

        Returns:
            (html.HtmlElement, dict): Metadata HTML root and its namespaces

        """
        if self.is_archived:
            root = files.read_archived_html(self.path, f".*{mtd_archived}")
        else:
            try:
                mtd_file = next(self.path.glob(f"**/*{mtd_from_path}"))
                if path.is_cloud_path(mtd_file):
                    try:
                        # Try using read_text (faster)
                        root = html.fromstring(mtd_file.read_text())
                    except ValueError:
                        # Try using read_bytes
                        # Slower but works with:
                        # {ValueError}Unicode strings with encoding declaration are not supported.
                        # Please use bytes input or XML fragments without declaration.
                        root = html.fromstring(mtd_file.read_bytes())
                else:
                    # pylint: disable=I1101:
                    # Module 'lxml.etree' has no 'parse' member, but source is unavailable.
                    html_tree = html.parse(str(mtd_file))
                    root = html_tree.getroot()
            except StopIteration as ex:
                raise InvalidProductError(
                    f"Metadata file ({mtd_from_path}) not found in {self.path}"
                ) from ex

        return root

    def read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S1A_IW_GRDH_1SDV_20191215T060906_20191215T060931_030355_0378F7_3696.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element product at 0x1832895d788>, '')

        Returns:
            (etree._Element, dict): Metadata XML root and its namespace
        """
        return self._read_mtd()

    # pylint: disable=W0613
    @abstractmethod
    def _read_band(
        self,
        band_path: AnyPathType,
        band: BandNames = None,
        pixel_size: Union[tuple, list, float] = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.DataArray:
        """
        Read band from disk.

        .. WARNING::
            For optical data, invalid pixels are not managed here

        Args:
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            pixel_size (Union[tuple, list, float]): Size of the pixels of the wanted band, in dataset unit (X, Y)
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            xr.DataArray: Band xarray

        """
        raise NotImplementedError

    def load(
        self,
        bands: Union[list, BandNames, str],
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.Dataset:
        """
        Open the bands and compute the wanted index.

        - For Optical data:
            The bands will be purged of nodata and invalid pixels (if specified with the CLEAN_OPTICAL keyword),
            the nodata will be set to -9999 and the bands will be DataArrays in float32.

        - For SAR data:
            The bands will be purged of nodata (not over the sea),
            the nodata will be set to 0 to respect SNAP's behavior and the bands will be DataArray in float32.

        Bands that come out this function at the same time are collocated and therefore have the same shapes.
        This can be broken if you load data separately. Its is best to always load DEM data with some real bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> bands = prod.load([GREEN, NDVI], pixel_size=20)

        Args:
            bands (Union[list, BandNames, str]): Band list
            pixel_size (float): Pixel size of the band, in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands

        Returns:
            xr.Dataset: Dataset with a variable per band
        """
        if not pixel_size and "resolution" in kwargs:
            logs.deprecation_warning(
                "`resolution` is deprecated in favor of `pixel_size` to avoid confusion. `resolution` will be removed in a future release."
            )
            pixel_size = kwargs.pop("resolution")

        if (isinstance(bands, list) and ("GREEN1" in bands or GREEN1 in bands)) or (
            "GREEN1" == bands or GREEN1 == bands
        ):
            logs.deprecation_warning(
                "`GREEN1` is deprecated in favor of `GREEN_1`. `GREEN1` will be removed in a future release."
            )

        # Check if all bands are valid
        bands = self.to_band(bands)

        for band in bands:
            assert self.has_band(band), f"{self.name} has not a {to_str(band)[0]} band."

        # Load bands (only once ! and convert the bands to be loaded to correct format)
        unique_bands = misc.unique(bands)
        band_xds = self._load(unique_bands, pixel_size, size, **kwargs)

        # Rename all bands and add attributes
        for key, val in band_xds.items():
            band_xds[key] = self._update_attrs(val, key, **kwargs)

        # Update stack's attributes
        if len(band_xds) > 0:
            band_xds = self._update_attrs(band_xds, bands, **kwargs)

        return band_xds

    def _load(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.Dataset:
        """
        Core function loading optical data bands

        Args:
            bands (list): Band list
            pixel_size (float): Pixel size of the band, in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands

        Returns:
            xr.Dataset: Dataset with a variable per band
        """
        band_list = []
        index_list = []
        dem_list = []
        clouds_list = []

        # Check if everything is valid
        for band in bands:
            if is_index(band):
                if self._has_index(band):
                    if band in indices.DEPRECATED_SPECTRAL_INDICES:
                        logs.deprecation_warning(
                            "Aliases of Awesome Spectral Indices won't be available in future versions of EOReader. "
                            f"Please use {indices.DEPRECATED_SPECTRAL_INDICES[band]} instead of {band}"
                        )
                    index_list.append(band)
                else:
                    raise InvalidIndexError(
                        f"{band} cannot be computed from {self.condensed_name}."
                    )
            elif is_sar_band(band):
                if self.sensor_type == SensorType.SAR:
                    if not self.has_band(band):
                        raise InvalidBandError(
                            f"{band} cannot be retrieved from {self.condensed_name}"
                        )
                    else:
                        band_list.append(band)
                else:
                    raise TypeError(
                        f"You should ask for Optical bands as {self.name} is an optical product."
                    )
            elif is_spectral_band(band):
                if self.sensor_type == SensorType.OPTICAL:
                    if self.has_band(band):
                        band_list.append(band)
                    else:
                        raise InvalidBandError(
                            f"{band} cannot be retrieved from {self.condensed_name}."
                        )
                else:
                    raise TypeError(
                        f"You should ask for SAR bands as {self.name} is a SAR product."
                    )
            elif is_dem(band):
                dem_list.append(band)
            elif is_clouds(band):
                if self.sensor_type == SensorType.OPTICAL:
                    clouds_list.append(band)
                else:
                    raise TypeError(
                        f"You cannot ask for cloud bands as {self.name} is a SAR product."
                    )

        # Check if DEM is set and exists
        if dem_list:
            self._check_dem_path(bands, **kwargs)

        # Get all bands to be open
        bands_to_load = band_list.copy()
        for idx in index_list:
            bands_to_load += NEEDED_BANDS[idx]

        # Load band arrays (only keep unique bands: open them only one time !)
        unique_bands = misc.unique(bands_to_load)
        bands_dict = {}
        if unique_bands:
            LOGGER.debug(f"Loading bands {to_str(unique_bands)}")
            loaded_bands = self._load_bands(
                unique_bands, pixel_size=pixel_size, size=size, **kwargs
            )

            # Compute index (they conserve the nodata)
            if index_list:
                # Collocate bands before indices to ensure the same size to perform operations between bands
                loaded_bands = self._collocate_bands(loaded_bands)

                LOGGER.debug(f"Loading indices {to_str(index_list)}")
                bands_dict.update(
                    self._load_spectral_indices(
                        index_list,
                        loaded_bands,
                        pixel_size=pixel_size,
                        size=size,
                        **kwargs,
                    )
                )

            # Add bands
            bands_dict.update({band: loaded_bands[band] for band in band_list})

        # Add DEM
        if dem_list:
            LOGGER.debug(f"Loading DEM bands {to_str(dem_list)}")
            bands_dict.update(
                self._load_dem(dem_list, pixel_size=pixel_size, size=size, **kwargs)
            )

        # Add Clouds
        if clouds_list:
            LOGGER.debug(f"Loading Cloud bands {to_str(clouds_list)}")
            bands_dict.update(
                self._load_clouds(
                    clouds_list, pixel_size=pixel_size, size=size, **kwargs
                )
            )

        # Manage the case of arrays of different size -> collocate arrays if needed
        bands_dict = self._collocate_bands(bands_dict)

        # Create a dataset (only after collocation)
        coords = None
        if bands_dict:
            coords = bands_dict[bands[0]].coords

        # Make sure the dataset has the bands in the right order -> re-order the input dict
        return xr.Dataset({key: bands_dict[key] for key in bands}, coords=coords)

    def _load_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        raise NotImplementedError

    @abstractmethod
    def _load_bands(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load bands as numpy arrays with the same pixel size (and same metadata).

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        raise NotImplementedError

    def _load_spectral_indices(
        self,
        index_list: list,
        loaded_bands: dict,
        pixel_size: Union[float, tuple] = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}
        for idx in index_list:
            idx_path = self._construct_band_path(
                idx, pixel_size, size, writable=False, **kwargs
            )
            if idx_path.is_file():
                band_dict[idx] = utils.read(idx_path)
            else:
                idx_arr = compute_index(index=idx, bands=loaded_bands).rename(idx)
                idx_arr.attrs["long_name"] = idx

                # Write on disk
                idx_path = self._construct_band_path(
                    idx, pixel_size, size, writable=True, **kwargs
                )
                utils.write(idx_arr, idx_path)
                band_dict[idx] = idx_arr

        return band_dict

    def _load_dem(
        self,
        band_list: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load bands as numpy arrays with the same pixel size (and same metadata).

        Args:
            band_list (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        dem_bands = {}
        if band_list:
            dem_path = os.environ.get(DEM_PATH)  # We already checked if it exists
            for band in band_list:
                assert is_dem(band)
                if band == DEM:
                    dem_path = self._warp_dem(
                        kwargs.get(DEM_KW, dem_path),
                        pixel_size=pixel_size,
                        size=size,
                        **kwargs,
                    )
                elif band == SLOPE:
                    dem_path = self._compute_slope(
                        kwargs.get(SLOPE_KW, dem_path),
                        pixel_size=pixel_size,
                        size=size,
                    )
                elif band == HILLSHADE:
                    dem_path = self._compute_hillshade(
                        kwargs.get(HILLSHADE_KW, dem_path),
                        pixel_size=pixel_size,
                        size=size,
                    )
                else:
                    raise InvalidTypeError(f"Unknown DEM band: {band}")

                dem_name = to_str(band)[0]
                dem_arr = utils.read(
                    dem_path, pixel_size=pixel_size, size=size, as_type=np.float32
                ).rename(dem_name)
                dem_arr.attrs["long_name"] = dem_name
                dem_bands[band] = dem_arr

        return dem_bands

    def has_band(self, band: Union[BandNames, str]) -> bool:
        """
        Does this product has the specified band ?

        By band, we mean:

        - satellite band
        - index
        - DEM band
        - cloud band

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.has_band(GREEN)
            True
            >>> prod.has_band(TIR_2)
            False
            >>> prod.has_band(NDVI)
            True
            >>> prod.has_band(SHADOWS)
            False
            >>> prod.has_band(HILLSHADE)
            True

        Args:
            band (Union[BandNames, str]): EOReader band (optical, SAR, clouds, DEM)

        Returns:
            bool: True if the products has the specified band
        """
        band = self.to_band(band)[0]

        if is_dem(band):
            if self.sensor_type == SensorType.SAR and band == HILLSHADE:
                has_band = False
            else:
                has_band = True
        elif is_clouds(band):
            has_band = self._has_cloud_band(band)
        elif is_index(band):
            has_band = self._has_index(band)
        else:
            has_band = band in self.get_existing_bands()

        return has_band

    def has_bands(self, bands: Union[list, BandNames, str]) -> bool:
        """
        Does this product has the specified bands ?

        By band, we mean:

        - satellite band
        - index
        - DEM band
        - cloud band

        See :code:`has_band` for a code example.

        Args:
            bands (Union[list, BandNames, str]): EOReader bands (optical, SAR, clouds, DEM)

        Returns:
            bool: True if the products has the specified band
        """
        if not isinstance(bands, list):
            bands = [bands]

        return all([self.has_band(band) for band in set(bands)])

    @abstractmethod
    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.has_cloud_band(CLOUDS)
            True
        """
        raise NotImplementedError

    def _has_index(self, idx: str) -> bool:
        """
        Cen the specified index be computed from this product ?

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.has_index(NDVI)
            True

        Args:
            idx (str): Index

        Returns:
            bool: True if the specified index can be computed with this product's bands
        """
        index_bands = to_band(indices.get_needed_bands(idx))
        return all(np.isin(index_bands, self.get_existing_bands()))

    def __gt__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired after the other

        """
        return self.date > other.date

    def __ge__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired after or in the same time as the other

        """
        return self.date >= other.date

    def __eq__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired in the same time as the other

        """
        return self.date == other.date

    def __ne__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired not in the same time as the other

        """
        return self.date != other.date

    def __le__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired before or in the same time as the other

        """
        return self.date <= other.date

    def __lt__(self, other: Product) -> bool:
        """
        Overload greater than for eoreader -> compare the dates:
        The greater products is the one acquired the last.

        Args:
            other (Product): Other products to be compared with this one

        Returns:
            bool: True if this product has been acquired before the other

        """
        return self.date < other.date

    def __hash__(self):
        return hash(self.condensed_name)

    def _get_out_path(self, filename: str) -> Tuple[AnyPathType, bool]:
        """
        Returns the output path of a file to be written, depending on if it already exists or not (manages CI folders)

        Args:
            filename (str): Filename

        Returns:
            Tuple[AnyPathType, bool]: Output path and if the file already exists or not
        """
        out = self._get_band_folder() / filename
        exists = True
        if not out.exists():
            exists = False
            out = self._get_band_folder(writable=True) / filename

        return out, exists

    @property
    def output(self) -> AnyPathType:
        """Output directory of the product, to write orthorectified data for example."""
        return self._output

    @output.setter
    def output(self, value: str):
        """Output directory of the product, to write orthorectified data for example."""
        # Set the new output
        self._output = AnyPath(value)
        if not path.is_cloud_path(self._output):
            self._output = self._output.resolve()

        # Create temporary process folder
        old_tmp_process = self._tmp_process
        self._tmp_process = self._output.joinpath(f"tmp_{self.condensed_name}")
        os.makedirs(self._tmp_process, exist_ok=True)

        # Move all files from old process folder into the new one
        for file in path.listdir_abspath(old_tmp_process):
            try:
                shutil.move(str(file), self._tmp_process)
            except shutil.Error:
                # Don't overwrite file
                pass

        # Remove old output if existing into the new output
        if self._tmp_output:
            self._tmp_output.cleanup()
            self._tmp_output = None

    @property
    def stac(self) -> StacItem:
        if not self._stac:
            self._stac = StacItem(self)

        return self._stac

    def _warp_dem(
        self,
        dem_path: str = "",
        pixel_size: Union[float, tuple] = None,
        size: Union[list, tuple] = None,
        resampling: Resampling = Resampling.bilinear,
        **kwargs,
    ) -> AnyPathType:
        """
        Get this product DEM, warped to this product footprint and CRS.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.warp_dem(pixel_size=20)  # In meters
            '/path/to/20200824T110631_S2_T30TTK_L1C_150432_DEM.tif'

        Args:
            dem_path (str): DEM path, using EUDEM/MERIT DEM if none
            pixel_size (Union[float, tuple]): Pixel size in meters. If not specified, use the product pixel size.
            resampling (Resampling): Resampling method
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands

        Returns:
            AnyPathType: DEM path (as a VRT)
        """
        dem_name = f"{self.condensed_name}_DEM_{path.get_filename(dem_path)}.vrt"

        warped_dem_path, warped_dem_exists = self._get_out_path(dem_name)
        if warped_dem_exists:
            LOGGER.debug(
                "Already existing DEM for %s. Skipping process.", self.condensed_name
            )
        else:
            LOGGER.debug("Warping DEM for %s", self.condensed_name)

            # Allow S3 HTTP Urls only on Linux because rasterio bugs on Windows
            if validators.url(dem_path) and platform.system() == "Windows":
                raise OSError(
                    f"URLs to DEM like {dem_path} are not supported on Windows! Use Docker or Linux instead"
                )

            # Check existence (SRTM)
            if not validators.url(dem_path):
                dem_path = AnyPath(dem_path)
                if not dem_path.is_file():
                    raise FileNotFoundError(f"DEM file does not exist here: {dem_path}")

            # Reproject DEM into products CRS
            LOGGER.debug("Using DEM: %s", dem_path)
            def_tr, def_w, def_h, def_crs = self.default_transform(**kwargs)
            with rasterio.open(str(dem_path)) as dem_ds:
                # Get adjusted transform and shape (with new pixel_size)
                if size is not None and pixel_size is None:
                    try:
                        # Get destination transform
                        out_h = size[1]
                        out_w = size[0]

                        # Get destination transform
                        coeff_x = def_w / out_w
                        coeff_y = def_h / out_h
                        dst_tr = def_tr
                        dst_tr *= dst_tr.scale(coeff_x, coeff_y)

                    except (TypeError, KeyError):
                        raise ValueError(
                            f"Size should exist (as pixel_size is None)"
                            f" and castable to a list: {size}"
                        )

                else:
                    # Refine pixel_size
                    if pixel_size is None:
                        pixel_size = self.pixel_size

                    bounds = transform.array_bounds(def_h, def_w, def_tr)
                    dst_tr, out_w, out_h = warp.calculate_default_transform(
                        def_crs,
                        self.crs(),
                        def_w,
                        def_h,
                        *bounds,
                        resolution=pixel_size,
                    )

                vrt_options = {
                    "resampling": resampling,
                    "crs": self.crs(),
                    "transform": dst_tr,
                    "height": out_h,
                    "width": out_w,
                    "nodata": self.nodata,
                    "dtype": "float32",
                }

                with WarpedVRT(dem_ds, **vrt_options) as vrt:
                    # At this point 'vrt' is a full dataset with dimensions,
                    # CRS, and spatial extent matching 'vrt_options'.
                    rio_shutil.copy(vrt, warped_dem_path, driver="vrt")

        return warped_dem_path

    @abstractmethod
    def _compute_hillshade(
        self,
        dem_path: str = "",
        pixel_size: Union[float, tuple] = None,
        size: Union[list, tuple] = None,
        resampling: Resampling = Resampling.bilinear,
    ) -> AnyPathType:
        """
        Compute Hillshade mask

        Args:
            dem_path (str): DEM path, using EUDEM/MERIT DEM if none
            pixel_size (Union[float, tuple]): Pixel size in meters. If not specified, use the product pixel size.
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            resampling (Resampling): Resampling method

        Returns:
            AnyPathType: Hillshade mask path

        """
        raise NotImplementedError

    def _compute_slope(
        self,
        dem_path: str = "",
        pixel_size: Union[float, tuple] = None,
        size: Union[list, tuple] = None,
        resampling: Resampling = Resampling.bilinear,
    ) -> AnyPathType:
        """
        Compute slope mask

        Args:
            dem_path (str): DEM path, using EUDEM/MERIT DEM if none
            pixel_size (Union[float, tuple]): Pixel size in meters. If not specified, use the product pixel size.
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            resampling (Resampling): Resampling method

        Returns:
            AnyPathType: Slope mask path

        """
        # Warp DEM
        warped_dem_path = self._warp_dem(dem_path, pixel_size, size, resampling)

        # Get slope path
        slope_name = f"{self.condensed_name}_SLOPE_{path.get_filename(dem_path)}.tif"

        slope_path, slope_exists = self._get_out_path(slope_name)
        if slope_exists:
            LOGGER.debug(
                "Already existing slope DEM for %s. Skipping process.",
                self.condensed_name,
            )
        else:
            LOGGER.debug("Computing slope for %s", self.condensed_name)

            # Compute slope
            slope = rasters.slope(warped_dem_path)
            utils.write(slope, slope_path)

        return slope_path

    @staticmethod
    def _collocate_bands(bands: dict, reference: xr.DataArray = None) -> dict:
        """
        Collocate all bands from a dict if needed (if a raster shape is different)

        Args:
            bands (dict): Dict of bands to collocate if needed
            reference (xr.DataArray): Reference array

        Returns:
            dict: Collocated bands
        """
        for band_id, band_arr in bands.items():
            if reference is None:
                # If reference is not passed, use the first array
                # Don't collocate if same array
                reference = band_arr
            else:
                # To be sure, always collocate arrays, even if the size is the same
                # Indeed, a small difference in the coordinates will lead to empy arrays
                # So the bands MUST BE exactly aligned
                bands[band_id] = rasters.collocate(reference, band_arr)

        return bands

    # pylint: disable=R0913
    # Too many arguments (6/5)
    def stack(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        stack_path: AnyPathStrType = None,
        save_as_int: bool = False,
        **kwargs,
    ) -> xr.DataArray:
        """
        Stack bands and index of a products.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> stack = prod.stack([NDVI, MNDWI, GREEN], pixel_size=20)  # In meters

        Args:
            bands (list): Bands and index combination
            pixel_size (float): Stack pixel size. . If not specified, use the product pixel size.
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            stack_path (AnyPathStrType): Stack path
            save_as_int (bool): Convert stack to uint16 to save disk space (and therefore multiply the values by 10.000)
            **kwargs: Other arguments passed to :code:`load` or :code:`rioxarray.to_raster()` (such as :code:`compress`)

        Returns:
            xr.DataArray: Stack as a DataArray
        """
        # Manage already existing stack on disk
        if stack_path:
            stack_path = AnyPath(stack_path)
            if stack_path.is_file():
                return utils.read(stack_path, resolution=pixel_size, size=size)
            else:
                os.makedirs(str(stack_path.parent), exist_ok=True)

        bands = self.to_band(bands)

        # Create the analysis stack
        band_xds = self.load(bands, pixel_size=pixel_size, size=size, **kwargs)

        # Stack bands
        if save_as_int:
            nodata = kwargs.pop("nodata", UINT16_NODATA)
        else:
            nodata = kwargs.pop("nodata", self.nodata)
        stack, dtype = utils.stack(band_xds, save_as_int, nodata, **kwargs)

        # Update stack's attributes
        stack = self._update_attrs(stack, bands, **kwargs)

        # Write on disk
        if stack_path:
            LOGGER.debug("Saving stack")
            utils.write(stack, stack_path, dtype=dtype, nodata=nodata, **kwargs)

        return stack

    @abstractmethod
    def _update_attrs_constellation_specific(
        self, xarr: xr.DataArray, bands: list, **kwargs
    ) -> xr.DataArray:
        """
        Update attributes of the given array (constellation specific)

        Args:
            xarr (xr.DataArray): Array whose attributes need an update
            bands (list): Array name (as a str or a list)

        Returns:
            xr.DataArray: Updated array/dataset
        """
        raise NotImplementedError

    def _update_attrs(
        self, xarr: Union[xr.DataArray, xr.Dataset], bands: list, **kwargs
    ) -> xr.DataArray:
        """
        Update attributes of the given array
        Args:
            xarr (Union[xr.DataArray, xr.Dataset]): Array whose attributes need an update
            bands (list): Bands
        Returns:
            xr.DataArray: Updated array
        """
        # Clean attributes, we don't want to pollute our attributes by default ones (not deterministic)
        # Are we sure of that ?
        xarr.attrs = {}

        if not isinstance(bands, list):
            bands = [bands]
        long_name = to_str(bands)
        xr_name = "_".join(long_name)
        attr_name = " ".join(long_name)

        if isinstance(xarr, xr.DataArray):
            xarr = xarr.rename(xr_name)
        xarr.attrs["long_name"] = attr_name
        xarr.attrs["constellation"] = (
            self.constellation
            if isinstance(self.constellation, str)
            else self.constellation.value
        )
        xarr.attrs["constellation_id"] = self.constellation_id
        xarr.attrs["product_path"] = str(self.path)  # Convert to string
        xarr.attrs["product_name"] = self.name
        xarr.attrs["product_filename"] = self.filename
        xarr.attrs["instrument"] = (
            self.instrument
            if isinstance(self.instrument, str)
            else self.instrument.value
        )
        xarr.attrs["product_type"] = (
            self.product_type
            if isinstance(self.product_type, str)
            else self.product_type.value
        )
        xarr.attrs["acquisition_date"] = self.get_datetime(as_datetime=False)
        xarr.attrs["condensed_name"] = self.condensed_name
        od = self.get_orbit_direction()
        xarr.attrs["orbit_direction"] = od.value if od is not None else str(od)

        # kwargs attrs
        xarr = self._update_attrs_constellation_specific(xarr, bands, **kwargs)

        return xarr

    @staticmethod
    def _check_dem_path(bands: list, **kwargs) -> None:
        """
        Check if DEM is set and exists if DEM bands are asked.

        Args:
            bands (list): List of the wanted bands
            kwargs: Other arguments used to load bands
        """
        if DEM_PATH not in os.environ:
            if (
                (DEM in bands and DEM_KW not in kwargs)
                or (SLOPE in bands and SLOPE_KW not in kwargs)
                or (HILLSHADE in bands and HILLSHADE_KW not in kwargs)
            ):
                raise ValueError(
                    f"DEM path not set, unable to compute DEM bands! "
                    f"Please set the environment variable {DEM_PATH} or a DEM keyword."
                )
        else:
            dem_path = os.environ.get(DEM_PATH)
            # URLs and file paths are required
            if not validators.url(dem_path):
                dem_path = AnyPath(dem_path)
                if not dem_path.is_file():
                    raise FileNotFoundError(
                        f"{dem_path} is not a file! "
                        f"Please set the environment variable {DEM_PATH} to an existing file."
                    )

    @cache
    def default_transform(self, **kwargs) -> (Affine, int, int, CRS):
        """
        Returns default transform data of the default band (UTM),
        as the :code:`rasterio.warp.calculate_default_transform` does:
        - transform
        - width
        - height
        - crs

        Args:
            kwargs: Additional arguments
        Returns:
            Affine, int, int, CRS: transform, width, height, CRS

        """
        with rasterio.open(str(self.get_default_band_path(**kwargs))) as dst:
            return dst.transform, dst.width, dst.height, dst.crs

    def _pixel_size_from_img_size(self, size: Union[list, tuple] = None) -> tuple:
        """
        Compute the corresponding pixel size to a given image size (positive resolution)

        Args:
            size (Union[list, tuple]): Size

        Returns:
            tuple: Pixel size as a tuple (x, y)
        """
        def_tr, def_w, def_h, def_crs = self.default_transform()
        bounds = transform.array_bounds(def_h, def_w, def_tr)

        # Manage WGS84 case
        if not def_crs.is_projected:
            utm_tr, utm_w, utm_h = warp.calculate_default_transform(
                def_crs,
                self.crs(),
                def_w,
                def_h,
                *bounds,
                resolution=self.pixel_size,
            )
            res_x = abs(utm_tr.a * utm_w / size[0])
            res_y = abs(utm_tr.e * utm_h / size[1])
        # Manage UTM case
        else:
            res_x = abs(def_tr.a * def_w / size[0])
            res_y = abs(def_tr.e * def_h / size[1])

        # Round pixel_size to the closest meter (under 1 meter, allow centimetric pixel_size)
        if res_x < 1.0:
            res_x = np.round(res_x, 1)
        else:
            res_x = np.round(res_x, 0)
        if res_y < 1.0:
            res_y = np.round(res_y, 1)
        else:
            res_y = np.round(res_y, 0)

        return res_x, res_y

    def clean_tmp(self):
        """
        Clean the temporary directory of the current product
        """
        if self._tmp_process.exists():
            for tmp_file in self._tmp_process.glob("*"):
                files.remove(tmp_file)

    def clear(self):
        """
        Clear this product's cache
        """
        # -- Delete all cached properties and functions
        gc.collect()

        # All objects collected
        objects = []
        for obj in gc.get_objects():
            try:
                if isinstance(obj, functools._lru_cache_wrapper):
                    objects.append(obj)
            except ReferenceError:
                pass

        # All objects cleared
        for obj in objects:
            obj.cache_clear()

    def _pixel_size_to_str(self, pixel_size: Union[float, tuple, list] = None):
        """
        Convert a pixel_size to a normalized string

        Args:
            pixel_size (Union[float, tuple, list]): Pixel size

        Returns:
            str: Pixel size as a string
        """

        def _res_to_str(res):
            return f"{abs(res):.2f}m".replace(".", "-")

        if pixel_size:
            if isinstance(pixel_size, (tuple, list)):
                res_x = _res_to_str(pixel_size[0])
                res_y = _res_to_str(pixel_size[1])
                if res_x == res_y:
                    res_str = res_x
                else:
                    res_str = f"{res_x}_{res_y}"
            else:
                res_str = _res_to_str(pixel_size)
        else:
            res_str = _res_to_str(self.pixel_size)

        return res_str

    def _to_repr(self) -> list:
        """
        Returns a representation of the product as a list

        Returns:
            list: Representation of the product
        """
        band_repr = "\n".join(
            [
                f"\t\t{band.value}: {val.id}"
                for band, val in self.bands.items()
                if val is not None
            ]
        )
        repr_str = [
            f"eoreader.{self.__class__.__name__} '{self.name}'",
            "Attributes:",
            f"\tcondensed_name: {self.condensed_name}",
            f"\tpath: {self.path}",
            f"\tconstellation: {self.constellation if isinstance(self.constellation, str) else self.constellation.value}",
            f"\tsensor type: {self.sensor_type if isinstance(self.sensor_type, str) else self.sensor_type.value}",
            f"\tproduct type: {self.product_type if isinstance(self.product_type, str) else self.product_type.value}",
            f"\tdefault pixel size: {self.pixel_size}",
            f"\tdefault resolution: {self.resolution}",
            f"\tacquisition datetime: {self.get_datetime(as_datetime=True).isoformat()}",
            f"\tband mapping:\n{band_repr}",
            f"\tneeds extraction: {self.needs_extraction}",
        ]

        return repr_str + self._to_repr_constellation_specific()

    @abstractmethod
    def _to_repr_constellation_specific(self) -> list:
        """
        Representation specific to the constellation

        Returns:
            list: Representation list (constellation specific)
        """
        raise NotImplementedError

    def __repr__(self):
        return "\n".join(self._to_repr())

    def get_quicklook_path(self) -> Union[None, str]:
        """
        Get quicklook path if existing (no such thing for Sentinel-2)

        Returns:
            str: Quicklook path
        """
        LOGGER.debug(f"No quicklook available for {self.constellation.value} data!")
        return None

    def plot(self) -> None:
        """
        Plot the quicklook if existing
        """
        try:
            import matplotlib.pyplot as plt
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "You need to install 'matplotlib' to plot the product."
            )
        else:
            quicklook_path = self.get_quicklook_path()

            if quicklook_path is not None:
                plt.figure(figsize=(6, 6))
                if path.get_ext(quicklook_path).lower() in ["png", "jpg", "jpeg"]:
                    try:
                        from PIL import Image
                    except ModuleNotFoundError:
                        raise ModuleNotFoundError(
                            "You need to install 'pillow' to plot the product."
                        )

                    if self.is_archived:
                        qlk = BytesIO(
                            files.read_archived_file(
                                self.path, f".*{os.path.basename(quicklook_path)}"
                            )
                        )
                    else:
                        if path.is_cloud_path(quicklook_path):
                            quicklook_path = AnyPath(quicklook_path).fspath

                        qlk = quicklook_path
                    plt.imshow(Image.open(qlk))
                else:
                    qck = rasters.read(quicklook_path)
                    if qck.rio.count == 3:
                        qck.plot.imshow(robust=True)
                    elif qck.rio.count == 1:
                        qck.plot(cmap="GnBu_r", robust=True)
                    else:
                        pass

                plt.title(f"{self.condensed_name}")

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        raise NotImplementedError

    def to_band(self, raw_bands: Union[list, BandNames, str, int]) -> list:
        """
        Convert any raw band identifier to a usable band.

        Bands can be called with their name, ID or mapped name.
        For example, for Sentinel-3 OLCI you can use `7`, `Oa07` or `YELLOW`. For Landsat-8, you can use `BLUE` or `2`.

        Args:
            raw_bands (Union[list, BandNames, str, int]): Raw bands

        Returns:
            list: Mapped bands
        """
        if not isinstance(raw_bands, list):
            raw_bands = [raw_bands]

        bands = []
        for raw_band in raw_bands:
            try:
                bands += to_band(raw_band)
            except InvalidTypeError:
                found = False
                for band_name, band in self.bands.items():
                    if band is not None and str(raw_band) in [
                        str(band.id),
                        band.name,
                        band.common_name,
                        band.eoreader_name,
                    ]:
                        bands.append(band_name)
                        found = True
                        break

                if not found:
                    raise InvalidTypeError(
                        f"Couldn't find any band in {self.condensed_name} that corresponds to {raw_band}."
                    )

        return bands



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RADARSAT-Constellation Mission products.
More info `here <https://www.asc-csa.gc.ca/eng/satellites/radarsat/technical-features/characteristics.asp>`_.
"""
import difflib
import logging
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from lxml import etree
from sertit import vectors
from sertit.misc import ListEnum
from sertit.vectors import WGS84

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class RcmProductType(ListEnum):
    """
    RADARSAT-Constellation projection identifier.
    Take a look `here <https://catalyst.earth/catalyst-system-files/help/references/gdb_r/RADARSAT-2.html>`_.
    """

    SLC = "SLC"
    """ Georeferenced. Single look complex data in slant range. """

    GRC = "GRC"
    """ Georeferenced. Multilook complex data in ground range. """

    GRD = "GRD"
    """ Georeferenced. Multilook detected data in ground range. """

    GCC = "GCC"
    """
    Geocoded to a map projection.
    Complex data.
    Data commonly projected to Universal Transverse Mercator Projection (UTM)
    or Universal Polar Stereographic (UPS) north of 84N or south of 80S.
    """

    GCD = "GCD"
    """
    Geocoded to a map projection.
    Detected data.
    Data commonly projected to Universal Transverse Mercator (UTM)
    or Universal Polar Stereographic (UPS) north of 84N or south of 80S.
    """


@unique
class RcmSensorMode(ListEnum):
    """
    RADARSAT-Constellation sensor mode.
    Take a look `here <https://www.asc-csa.gc.ca/eng/satellites/radarsat/technical-features/characteristics.asp>`_.

    .. WARNING:: The name in the metadata may vary !
    """

    THREE_M = "Very High Resolution 3m"
    """" Very-High Resolution, 3 meters [3M] """

    FIVE_M = "High Resolution 5m"
    """" High Resolution, 5 meters [5M] """

    QP = "Quad-Polarization"
    """" Quad-Polarization [QP] """

    SIXTEEN_M = "Medium Resolution 16m"
    """" Medium Resolution, 16 meters [16M] """

    THIRTY_M = "Medium Resolution 30m"
    """" Medium Resolution, 30 meters [SC30] """

    FIFTY_M = "Medium Resolution 50m"
    """" Medium Resolution, 50 meters [SC50] """

    SCLN = "Low Noise"
    """" Low Noise [SCLN] """

    HUNDRED_M = "Low Resolution 100m"
    """ Low Resolution, 100 meters [SC100] """

    SD = "Ship Detection"
    """ Ship Detection Mode """

    FSL = "Spotlight Mode"
    """ Spotlight Mode [FSL] """


class RcmProduct(SarProduct):
    """
    Class for RADARSAT-Constellation Products

    You can use directly the .zip file
    """

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://ftp.maps.canada.ca/pub/csa_asc/Space-technology_Technologie-spatiale/radarsat_constellation_mission_plan/RCM-SP-52-9092_Product_Spec_1-15_Public.pdf>`_
        for more information (Table 4-1 Raw and Image Product Description Table - Contents and Volumes, page 18)
        """
        if self.sensor_mode == RcmSensorMode.THREE_M:
            def_pixel_size = 1.25
            def_res = 3.0
        elif self.sensor_mode == RcmSensorMode.FIVE_M:
            def_pixel_size = 2.0
            def_res = 5.0
        elif self.sensor_mode == RcmSensorMode.QP:
            def_pixel_size = 2.5
            def_res = 9.0
        elif self.sensor_mode == RcmSensorMode.SIXTEEN_M:
            def_pixel_size = 6.25
            def_res = 16.0
        elif self.sensor_mode == RcmSensorMode.THIRTY_M:
            def_pixel_size = 12.5
            def_res = 30.0
        elif self.sensor_mode == RcmSensorMode.FIFTY_M:
            def_pixel_size = 20.0
            def_res = 50.0
        elif self.sensor_mode in [RcmSensorMode.HUNDRED_M, RcmSensorMode.SCLN]:
            def_pixel_size = 40.0
            def_res = 100.0
        elif self.sensor_mode == RcmSensorMode.FSL:
            def_pixel_size = 0.33
            def_res = 1.0  # 3.0*1.0
        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        self.pixel_size = def_pixel_size
        self.resolution = def_res

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # SNAP cannot process zipped RCM
        self.needs_extraction = True

        # Its original filename is its name
        self._use_filename = True

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """
        # Private attributes
        self._raw_band_regex = "*_{}.tif"
        self._band_folder = self.path / "imagery"
        self.snap_filename = ""

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"RS2_OK73950_PK661843_DK590667_U25W2_20160228_112418_HH_SGF.zip"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent
                                                        geometry
            1  POLYGON ((106.57999 -6.47363, 107.06926 -6.473...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        # Open extent KML file
        try:
            extent_file = next(self.path.joinpath("preview").glob("*mapOverlay.kml"))
            product_kml = vectors.read(extent_file)
        except IndexError as ex:
            raise InvalidProductError(
                f"Extent file (product.kml) not found in {self.path}"
            ) from ex

        extent_wgs84 = product_kml[
            product_kml.Name == "Polygon Outline"
        ].envelope.to_crs(WGS84)

        return gpd.GeoDataFrame(geometry=extent_wgs84.geometry, crs=extent_wgs84.crs)

    def _set_instrument(self) -> None:
        """
        Set instrument

        RCM: https://earth.esa.int/web/eoportal/satellite-missions/r/rcm
        """
        self.instrument = "SAR C-band"

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Open identifier
        prod_type = root.findtext(f".//{namespace}productType")
        if not prod_type:
            raise InvalidProductError("productType not found in metadata!")

        self.product_type = RcmProductType.from_value(prod_type)

        if self.product_type in [RcmProductType.GRD, RcmProductType.GCD]:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type in [
            RcmProductType.SLC,
            RcmProductType.GRC,
            RcmProductType.GCC,
        ]:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not available for {self.name}"
            )

        if self.product_type != RcmProductType.GRD:
            LOGGER.warning(
                "Other products type than SGF has not been tested for %s data. "
                "Use it at your own risks !",
                self.constellation.value,
            )

    def _set_sensor_mode(self) -> None:
        """
        Get products type from RADARSAT-2 products name (could check the metadata too)
        """
        # Get metadata
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Get sensor mode
        # WARNING: this word may differ from the Enum !!! (no docs available)
        # Get the closest match
        sensor_mode_xml = root.findtext(f".//{namespace}acquisitionType")

        if sensor_mode_xml:
            sensor_mode = difflib.get_close_matches(
                sensor_mode_xml, RcmSensorMode.list_values()
            )[0]
            try:
                self.sensor_mode = RcmSensorMode.from_value(sensor_mode)
            except ValueError as ex:
                raise InvalidTypeError(f"Invalid sensor mode for {self.name}") from ex
        else:
            raise InvalidTypeError(f"Invalid sensor mode for {self.name}")

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"RS2_OK73950_PK661843_DK590667_U25W2_20160228_112418_HH_SGF.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2016, 2, 28, 11, 24, 18)
            >>> prod.get_datetime(as_datetime=False)
            '20160228T112418'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Open identifier
        acq_date = root.findtext(f".//{namespace}rawDataStartTime")
        if not acq_date:
            raise InvalidProductError("rawDataStartTime not found in metadata!")

        # Convert to datetime
        date = datetime.strptime(acq_date, "%Y-%m-%dT%H:%M:%S.%fZ")

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # The name is not in the classic metadata, but can be found in the manifest
        try:
            mtd_from_path = "preview/productPreview.html"
            mtd_archived = r"preview.*productPreview\.html"

            root = self._read_mtd_html(mtd_from_path, mtd_archived)

            # Open identifier
            name = root.findtext(".//header/h2")
            if not name:
                raise InvalidProductError("header/h2 not found in metadata!")

        except InvalidProductError:
            LOGGER.warning(
                "productPreview.html not found in the product, the name will be the filename"
            )
            name = self.filename

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"LC08_L1GT_023030_20200518_20200527_01_T2"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {http://www.rsi.ca/rs2/prod/xml/schemas}product at 0x1c0efbd37c8>,
            {None: '{http://www.rsi.ca/rs2/prod/xml/schemas}'})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespace
        """
        mtd_from_path = "metadata/product.xml"
        mtd_archived = r"metadata.*product\.xml"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def _get_condensed_name(self) -> str:
        """
        Get products condensed name ({acq_datetime}_{constellation}_{polarization}_{sensor_mode}_{product_type}).

        Returns:
            str: Condensed RCM name
        """
        # Get back the correct sensor mode name
        if self.sensor_mode == RcmSensorMode.THREE_M:
            mode_name = "3M"
        elif self.sensor_mode == RcmSensorMode.FIVE_M:
            mode_name = "5M"
        elif self.sensor_mode == RcmSensorMode.SIXTEEN_M:
            mode_name = "16M"
        elif self.sensor_mode == RcmSensorMode.THIRTY_M:
            mode_name = "SC30"
        elif self.sensor_mode == RcmSensorMode.FIFTY_M:
            mode_name = "SC50"
        elif self.sensor_mode == RcmSensorMode.HUNDRED_M:
            mode_name = "SC100"
        else:
            mode_name = self.sensor_mode.name

        pol_chan = [pol.value for pol in self.pol_channels]
        return f"{self.get_datetime()}_{self.constellation.name}_{'_'.join(pol_chan)}_{mode_name}_{self.product_type.value}"

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            quicklook_path = str(next(self.path.glob("preview/productOverview.png")))
        except StopIteration:
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Get the orbit direction
        try:
            od = OrbitDirection.from_value(
                root.findtext(f".//{namespace}passDirection").upper()
            )

        except TypeError:
            raise InvalidProductError("passDirection not found in metadata!")

        return od



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Band module containing:

- wrapper for SAR and optical bands
- Index definitions
- Aliases for all these useful variables

To use it, simply type:

.. code-block:: python

    >>> from eoreader.bands import *
    >>> GREEN
    <SpectralBandNames.GREEN: 'GREEN'>
    >>> HH
    <SarBandNames.HH: 'HH'>
    >>> NDVI
    <function NDVI at 0x00000261F6FFA950>
"""
from typing import Any, Union

# flake8: noqa
from eoreader.bands.bands import Band, BandMap

__all__ = ["Band", "BandMap"]

from eoreader.bands.band_names import BandNames

__all__ += ["BandNames"]

from eoreader.bands.band_names import (
    ALL_CLOUDS,
    CIRRUS,
    CLOUDS,
    RAW_CLOUDS,
    SHADOWS,
    CloudsBandNames,
)

__all__ += [
    "CloudsBandNames",
    "RAW_CLOUDS",
    "CLOUDS",
    "SHADOWS",
    "CIRRUS",
    "ALL_CLOUDS",
]

from eoreader.bands.band_names import DEM, HILLSHADE, SLOPE, DemBandNames

__all__ += ["DemBandNames", "DEM", "SLOPE", "HILLSHADE"]

from eoreader.bands.indices import (
    get_all_index_names,
    get_all_needed_bands,
    get_needed_bands,
    is_index,
)

__all__ += [
    "get_all_index_names",
    "get_needed_bands",
    "get_all_needed_bands",
    "is_index",
]

# Spyndex indices
__all__ += get_all_index_names()
from eoreader.bands.band_names import (
    HH,
    HH_DSPK,
    HV,
    HV_DSPK,
    RH,
    RH_DSPK,
    RV,
    RV_DSPK,
    VH,
    VH_DSPK,
    VV,
    VV_DSPK,
    SarBandNames,
)
from eoreader.bands.indices import *
from eoreader.bands.sar_bands import SarBand, SarBandMap

__all__ += [
    "SarBand",
    "SarBandMap",
    "SarBandNames",
    "VV",
    "VV_DSPK",
    "HH",
    "HH_DSPK",
    "VH",
    "VH_DSPK",
    "HV",
    "HV_DSPK",
    "RH",
    "RH_DSPK",
    "RV",
    "RV_DSPK",
]

from eoreader.bands.band_names import GREEN1  # To be deprecated
from eoreader.bands.band_names import (
    BLUE,
    CA,
    F1,
    F2,
    GREEN,
    GREEN_1,
    NARROW_NIR,
    NIR,
    PAN,
    RED,
    S7,
    SWIR_1,
    SWIR_2,
    SWIR_CIRRUS,
    TIR_1,
    TIR_2,
    VRE_1,
    VRE_2,
    VRE_3,
    WV,
    YELLOW,
    Oa01,
    Oa02,
    Oa09,
    Oa10,
    Oa13,
    Oa14,
    Oa15,
    Oa18,
    Oa19,
    Oa21,
    SpectralBandNames,
)
from eoreader.bands.spectral_bands import SpectralBand, SpectralBandMap

__all__ += [
    "SpectralBand",
    "SpectralBandMap",
    "SpectralBandNames",
    "CA",
    "BLUE",
    "GREEN",
    "GREEN_1",
    "GREEN1",  # To be deprecated
    "YELLOW",
    "RED",
    "VRE_1",
    "VRE_2",
    "VRE_3",
    "NIR",
    "NARROW_NIR",
    "WV",
    "SWIR_CIRRUS",
    "SWIR_1",
    "SWIR_2",
    "TIR_1",
    "TIR_2",
    "PAN",
    "S7",
    "F1",
    "F2",
    "Oa01",
    "Oa02",
    "Oa09",
    "Oa10",
    "Oa13",
    "Oa14",
    "Oa15",
    "Oa18",
    "Oa19",
    "Oa21",
]

from eoreader.bands.mappings import (
    EOREADER_STAC_MAP,
    EOREADER_TO_SPYNDEX_DICT,
    SPYNDEX_TO_EOREADER_DICT,
)

__all__ += ["EOREADER_TO_SPYNDEX_DICT", "SPYNDEX_TO_EOREADER_DICT", "EOREADER_STAC_MAP"]

__all__ += [
    "is_spectral_band",
    "is_thermal_band",
    "is_sar_band",
    "is_sat_band",
    "is_clouds",
    "is_dem",
    "to_band",
    "to_str",
]

from eoreader.exceptions import InvalidTypeError as _ite


def is_spectral_band(band: Any) -> bool:
    """
    Returns True if is a spectral band (from :code:`SpectralBandNames`)

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_spectral_band(NDVI)
        False
        >>> is_spectral_band(HH)
        False
        >>> is_spectral_band(GREEN)
        True
        >>> is_spectral_band(SLOPE)
        False
        >>> is_spectral_band(CLOUDS)
        False

    Args:
        band (Any): Anything that could be an optical band

    Returns:
        bool: True if the band asked is an optical band

    """
    is_valid = True
    try:
        SpectralBandNames(band)
    except ValueError:
        is_valid = False
    return is_valid


def is_thermal_band(band: Any) -> bool:
    """
    Returns True if is a spectral and a thermal band (from :code:`SpectralBandNames`)

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_spectral_band(NDVI)
        False
        >>> is_spectral_band(HH)
        False
        >>> is_spectral_band(GREEN)
        True
        >>> is_spectral_band(SLOPE)
        False
        >>> is_spectral_band(CLOUDS)
        False

    Args:
        band (Any): Anything that could be an optical band

    Returns:
        bool: True if the band asked is an optical band

    """
    return is_spectral_band(band) and band in [TIR_1, TIR_2, F1, F2, S7]


def is_sar_band(band: Any) -> bool:
    """
    Returns True if is a SAR band (from :code:`SarBandNames`)

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_sar_band(NDVI)
        False
        >>> is_sar_band(HH)
        True
        >>> is_sar_band(GREEN)
        False
        >>> is_sar_band(SLOPE)
        False
        >>> is_sar_band(CLOUDS)
        False

    Args:
        band (Any): Anything that could be a SAR band

    Returns:
        bool: True if the band asked is a SAR band

    """
    is_valid = True
    try:
        SarBandNames(band)
    except ValueError:
        is_valid = False
    return is_valid


def is_sat_band(band) -> bool:
    """
    Returns True if is a band (from both :code:`SarBandNames` or :code:`SpectralBandNames`)

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_sat_band(NDVI)
        False
        >>> is_sat_band(HH)
        True
        >>> is_sat_band(GREEN)
        True
        >>> is_sat_band(SLOPE)
        False
        >>> is_sat_band(CLOUDS)
        False

    Args:
        band (Any): Anything that could be a band

    Returns:
        bool: True if the band asked is a band

    """
    return is_sar_band(band) or is_spectral_band(band)


def is_clouds(clouds: Any) -> bool:
    """
    Returns True if we have a Clouds-related keyword

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_clouds(NDVI)
        False
        >>> is_clouds(HH)
        False
        >>> is_clouds(GREEN)
        False
        >>> is_clouds(SLOPE)
        False
        >>> is_clouds(CLOUDS)
        True
    """
    is_valid = True
    try:
        CloudsBandNames(clouds)
    except ValueError:
        is_valid = False
    return is_valid


def is_dem(dem: Any) -> bool:
    """
    Returns True if we have a DEM-related keyword

    .. code-block:: python

        >>> from eoreader.bands import *
        >>> is_dem(NDVI)
        False
        >>> is_dem(HH)
        False
        >>> is_dem(GREEN)
        False
        >>> is_dem(SLOPE)
        True
        >>> is_dem(CLOUDS)
        False
    """
    is_valid = True
    try:
        DemBandNames(dem)
    except ValueError:
        is_valid = False
    return is_valid


def to_band(to_convert: Union[list, BandNames, str]) -> list:
    """
    Convert a string (or real value) to any alias, band or index.

    You can pass the name or the value of the bands.

    .. code-block:: python

        >>> to_band(["NDVI", "GREEN", RED, "VH_DSPK", "SLOPE", DEM, "CLOUDS", CLOUDS])
        [<function NDVI at 0x00000154DDB12488>,
        <SpectralBandNames.GREEN: 'GREEN'>,
        <SpectralBandNames.RED: 'RED'>,
        <SarBandNames.VH_DSPK: 'VH_DSPK'>,
        <DemBandNames.SLOPE: 'SLOPE'>,
        <DemBandNames.DEM: 'DEM'>,
        <ClassifBandNames.CLOUDS: 'CLOUDS'>,
        <ClassifBandNames.CLOUDS: 'CLOUDS'>]

    Args:
        to_convert (Union[list, BandNames, str]): Values to convert into band objects

    Returns:
        list: converted values

    """
    if not isinstance(to_convert, list):
        to_convert = [to_convert]

    band_list = []
    for tc in to_convert:
        band_or_idx = None
        # Try legit types
        if isinstance(tc, str):
            # Try index
            if tc in get_all_index_names():
                from eoreader.bands import indices

                band_or_idx = getattr(indices, tc)
            else:
                try:
                    band_or_idx = SarBandNames.convert_from(tc)[0]
                except TypeError:
                    try:
                        band_or_idx = SpectralBandNames.convert_from(tc)[0]
                    except TypeError:
                        try:
                            band_or_idx = DemBandNames.convert_from(tc)[0]
                        except TypeError:
                            try:
                                band_or_idx = CloudsBandNames.convert_from(tc)[0]
                            except TypeError:
                                pass

        elif is_index(tc) or is_sat_band(tc) or is_dem(tc) or is_clouds(tc):
            band_or_idx = tc

        # Store it
        if band_or_idx:
            band_list.append(band_or_idx)
        else:
            if tc == "GREEN1":
                from sertit import logs

                logs.deprecation_warning(
                    "`GREEN1` is deprecated in favor of `GREEN_1`. `GREEN1` will be removed in a future release."
                )
                band_list.append(GREEN_1)
            else:
                raise _ite(f"Unknown band or index: {tc}")

    return band_list


def to_str(to_convert: Union[list, BandNames, str]) -> list:
    """
    Convert a string (or real value) to any alias, band or index.

    You can pass the name or the value of the bands.

    .. code-block:: python

        >>> to_str(["NDVI", "GREEN", RED, "VH_DSPK", "SLOPE", DEM, "CLOUDS", CLOUDS])
        ['NDVI', 'GREEN', 'RED', 'VH_DSPK', 'SLOPE', 'DEM', 'CLOUDS', 'CLOUDS']

    Args:
        to_convert (Union[list, BandNames, str]): Values to convert into str

    Returns:
        list: str bands
    """
    if not isinstance(to_convert, list):
        to_convert = [to_convert]

    bands_str = []
    for tc in to_convert:
        if isinstance(tc, str):
            band_str = tc
        else:
            try:
                band_str = tc.name
            except AttributeError:
                band_str = tc.__name__

        bands_str.append(band_str)

    return bands_str



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
ICEYE products.
Take a look
`here <https://www.iceye.com/hubfs/Downloadables/ICEYE-Level-1-Product-Specs-2019.pdf>`_.
"""
import logging
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from lxml import etree
from sertit import path, vectors
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import SarBandNames as sab
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.keywords import ICEYE_USE_SLC
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class IceyeProductType(ListEnum):
    """
    ICEYE products types.
    Take a look
    `here <https://www.iceye.com/hubfs/Downloadables/ICEYE-Level-1-Product-Specs-2019.pdf>`_.
    """

    GRD = "GRD"
    """ Level-1 Ground Range Detected (GRD) """

    SLC = "SLC"
    """ Level-1 Single Look Complex (SLC) """

    # Not released yet
    # ORTHO = "ORTHO"
    # """ ORTHO """


@unique
class IceyeSensorMode(ListEnum):
    """
    ICEYE imaging mode.
    Take a look
    `here <https://www.iceye.com/hubfs/Downloadables/ICEYE-Level-1-Product-Specs-2019.pdf>`_.
    """

    SM = "stripmap"
    """SM - Stripmap, StripmapWide, StripmapHI"""

    SL = "spotlight"
    """SL - Spotlight"""

    SC = "scan"
    """SC - ScanSAR"""


class IceyeProduct(SarProduct):
    """
    Class for ICEYE Products
    Take a look
    `here <https://www.iceye.com/hubfs/Downloadables/ICEYE-Level-1-Product-Specs-2019.pdf>`_.
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self._use_slc = kwargs.pop(ICEYE_USE_SLC, None)

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://sar.iceye.com/5.0/productguide/collectioncharacteristics/>`_ for more information (Amplitude Image Parameters)
        """
        if self.sensor_mode == IceyeSensorMode.SM:
            def_pixel_size = 2.5
            def_res = 3.0
        elif self.sensor_mode == IceyeSensorMode.SL:
            def_pixel_size = 0.5
            def_res = 1.0
        elif self.sensor_mode == IceyeSensorMode.SC:
            def_pixel_size = 6.0
            def_res = 15.0

        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        self.pixel_size = def_pixel_size
        self.resolution = def_res

    def _set_instrument(self) -> None:
        """
        Set instrument

        ICEYE: https://earth.esa.int/eogateway/missions/iceye
        """
        self.instrument = "SAR X-band"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._band_folder = self.path

        # SNAP cannot process its archive
        self.needs_extraction = True

        # Its original filename is its name
        self._use_filename = True

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """
        # Private attributes
        if self._use_slc:
            self.snap_filename = str(next(self.path.glob("*ICEYE*SLC*.xml")).name)
            self._raw_band_regex = "*ICEYE*SLC*.h5"
        else:
            self.snap_filename = str(next(self.path.glob("*ICEYE*GRD*.xml")).name)
            self._raw_band_regex = "*ICEYE*GRD*.tif"

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                                        geometry
            0  POLYGON ((108.09797 15.61011, 108.48224 15.678...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        # Open extent KML file
        try:
            extent_file = next(self.path.glob("*ICEYE*QUICKLOOK*.kml"))
        except IndexError as ex:
            raise InvalidProductError(
                f"Extent file (*.kml) not found in {self.path}"
            ) from ex

        extent_wgs84 = vectors.read(extent_file).envelope

        return gpd.GeoDataFrame(geometry=extent_wgs84.geometry, crs=extent_wgs84.crs)

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Some ICEYE product metadata has a namespace some don't
        namespace = nsmap.get(None, "")

        # Open identifier
        prod_type = root.findtext(f".//{namespace}product_level")
        if not prod_type:
            raise InvalidProductError("mode not found in metadata!")

        self.product_type = IceyeProductType.from_value(prod_type)

        if self.product_type == IceyeProductType.GRD:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type == IceyeProductType.SLC:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not available for {self.name}"
            )

    def _set_sensor_mode(self) -> None:
        """Get sensor mode"""
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Some ICEYE product metadata has a namespace some don't
        namespace = nsmap.get(None, "")

        # Open identifier
        imaging_mode = root.findtext(f".//{namespace}acquisition_mode")
        if not imaging_mode:
            raise InvalidProductError("imagingMode not found in metadata!")

        # Get sensor mode
        try:
            self.sensor_mode = IceyeSensorMode.from_value(imaging_mode)
        except ValueError as ex:
            raise InvalidTypeError(f"Invalid sensor mode for {self.name}") from ex

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 10, 28, 22, 46, 25)
            >>> prod.get_datetime(as_datetime=False)
            '20201028T224625'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, nsmap = self.read_mtd()

            # Some ICEYE product metadata has a namespace some don't
            namespace = nsmap.get(None, "")

            # Open identifier
            acq_date = root.findtext(f".//{namespace}acquisition_start_utc")
            if not acq_date:
                raise InvalidProductError(
                    "acquisition_start_utc not found in metadata!"
                )

            # Convert to datetime
            date = datetime.strptime(acq_date, "%Y-%m-%dT%H:%M:%S.%f")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Some ICEYE product metadata has a namespace some don't
        namespace = nsmap.get(None, "")

        # Open identifier
        name = root.findtext(f".//{namespace}product_name")
        if not name:
            raise InvalidProductError("product_name not found in metadata!")

        # Check if `use_slc` is compatible
        if self._use_slc and IceyeProductType.SLC.value not in name:
            raise InvalidProductError(f"This product {self.name} has no SLC image!")
        elif not self._use_slc and IceyeProductType.GRD.value not in name:
            raise InvalidProductError(f"This product {self.name} has no GRD image!")

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1001513-735093"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element DeliveryNote at 0x2454ad4ee88>, {})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """

        def __read_mtd(prod_type: IceyeProductType):
            return self._read_mtd_xml(f"ICEYE*{prod_type.value}*.xml")

        if self._use_slc is None:
            try:
                root, nsmap = __read_mtd(IceyeProductType.GRD)
                self._use_slc = False
            except InvalidProductError:
                root, nsmap = __read_mtd(IceyeProductType.SLC)
                self._use_slc = True
        else:
            if self._use_slc:
                try:
                    root, nsmap = __read_mtd(IceyeProductType.SLC)
                except InvalidProductError:
                    LOGGER.warning("SLC image is not available for this product.")
                    self._use_slc = False
                    root, nsmap = __read_mtd(IceyeProductType.GRD)
            else:
                try:
                    root, nsmap = __read_mtd(IceyeProductType.GRD)
                except InvalidProductError:
                    LOGGER.warning("GRD image is not available for this product.")
                    self._use_slc = True
                    root, nsmap = __read_mtd(IceyeProductType.SLC)

        return root, nsmap

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the existing path of the VV band (as they come with the archived products).
        ICEYE product only contains a VV band !

        Args:
            **kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of every band existing in the raw products
        """
        band_paths = {}
        try:
            band_paths[sab.VV] = path.get_file_in_dir(
                self._band_folder, self._raw_band_regex, exact_name=True, get_list=False
            )
        except FileNotFoundError:
            raise InvalidProductError(
                "An ICEYE product should at least contain a VV band !"
            )

        return band_paths

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            quicklook_path = str(next(self.path.glob("*.png")))
        except StopIteration:
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Get the orbit direction
        try:
            od = OrbitDirection.from_value(root.findtext(".//orbit_direction"))

        except TypeError:
            raise InvalidProductError("orbit_direction not found in metadata!")

        return od



from eoreader.bands.band_names import (
    BLUE,
    CA,
    EOREADER_STAC_MAP,
    GREEN,
    HH,
    HV,
    NARROW_NIR,
    NIR,
    RED,
    SWIR_1,
    SWIR_2,
    TIR_1,
    TIR_2,
    VH,
    VRE_1,
    VRE_2,
    VRE_3,
    VV,
    WV,
    YELLOW,
)

EOREADER_TO_SPYNDEX_DICT = {
    CA: "A",
    BLUE: "B",
    GREEN: "G",
    RED: "R",
    VRE_1: "RE1",
    VRE_2: "RE2",
    VRE_3: "RE3",
    NIR: "N",
    NARROW_NIR: "N2",
    SWIR_1: "S1",
    SWIR_2: "S2",
    TIR_1: "T1",
    TIR_2: "T2",
    WV: "WV",
    YELLOW: "Y",
    VV: "VV",
    VH: "VH",
    HH: "HH",
    HV: "HV",
}

SPYNDEX_TO_EOREADER_DICT = dict(
    zip(EOREADER_TO_SPYNDEX_DICT.values(), EOREADER_TO_SPYNDEX_DICT.keys())
)

EOREADER_STAC_MAP = EOREADER_STAC_MAP



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Maxar satellites (GeoEye, WorldViews...) class.
See `here <https://earth.esa.int/eogateway/documents/20142/37627/DigitalGlobe-Standard-Imagery.pdf>`_
for more information.
"""
import logging
from abc import abstractmethod
from collections import namedtuple
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
import numpy as np
import xarray as xr
from lxml import etree
from rasterio import crs as riocrs
from sertit import geometry, path, rasters, vectors
from sertit.misc import ListEnum
from sertit.types import AnyPathType
from shapely.geometry import Polygon

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import (
    BLUE,
    CA,
    GREEN,
    NARROW_NIR,
    NIR,
    PAN,
    RED,
    VRE_1,
    VRE_2,
    VRE_3,
    WV,
    YELLOW,
    BandNames,
    SpectralBand,
)
from eoreader.exceptions import InvalidProductError
from eoreader.products import VhrProduct
from eoreader.products.optical.optical_product import RawUnits
from eoreader.reader import Constellation
from eoreader.stac import GSD, ID, NAME, WV_MAX, WV_MIN
from eoreader.utils import simplify

LOGGER = logging.getLogger(EOREADER_NAME)

_MAXAR_BAND_MTD = {
    NIR: "N",
    NARROW_NIR: "N",
    RED: "R",
    GREEN: "G",
    BLUE: "B",
    WV: "N2",
    VRE_1: "RE",
    VRE_2: "RE",
    VRE_3: "RE",
    YELLOW: "Y",
    CA: "C",
    PAN: "P",
}

GainOffset = namedtuple("GainOffset", ["gain", "offset"], defaults=[1.0, 0.0])
_MAXAR_GAIN_OFFSET = {
    Constellation.GE01: {
        PAN: GainOffset(gain=1.001, offset=0.0),
        BLUE: GainOffset(gain=1.041, offset=0.0),
        GREEN: GainOffset(gain=0.972, offset=0.0),
        RED: GainOffset(gain=0.979, offset=0.0),
        NIR: GainOffset(gain=0.951, offset=0.0),
        NARROW_NIR: GainOffset(gain=0.951, offset=0.0),
    },  # 2018v0
    Constellation.WV02: {
        PAN: GainOffset(gain=0.949, offset=-5.523),
        CA: GainOffset(gain=1.203, offset=-11.839),
        BLUE: GainOffset(gain=1.002, offset=-9.835),
        GREEN: GainOffset(gain=0.953, offset=-7.218),
        YELLOW: GainOffset(gain=0.946, offset=-5.675),
        RED: GainOffset(gain=0.955, offset=-5.046),
        VRE_1: GainOffset(gain=0.980, offset=-6.114),
        VRE_2: GainOffset(gain=0.980, offset=-6.114),
        VRE_3: GainOffset(gain=0.980, offset=-6.114),
        NIR: GainOffset(gain=0.966, offset=-5.096),
        NARROW_NIR: GainOffset(gain=0.966, offset=-5.096),
        WV: GainOffset(gain=1.01, offset=-4.059),
    },  # 2018v0
    Constellation.WV03: {
        PAN: GainOffset(gain=0.955, offset=-5.505),
        CA: GainOffset(gain=0.938, offset=-13.099),
        BLUE: GainOffset(gain=0.946, offset=-9.409),
        GREEN: GainOffset(gain=0.958, offset=-7.771),
        YELLOW: GainOffset(gain=0.979, offset=-5.489),
        RED: GainOffset(gain=0.969, offset=-4.579),
        VRE_1: GainOffset(gain=1.027, offset=-5.552),
        VRE_2: GainOffset(gain=1.027, offset=-5.552),
        VRE_3: GainOffset(gain=1.027, offset=-5.552),
        NIR: GainOffset(gain=0.977, offset=-6.508),
        NARROW_NIR: GainOffset(gain=0.977, offset=-6.508),
        WV: GainOffset(gain=1.007, offset=-3.699),
    },  # 2018v0
    Constellation.WV04: {
        PAN: GainOffset(gain=1.0, offset=0.0),
        BLUE: GainOffset(gain=1.0, offset=0.0),
        GREEN: GainOffset(gain=1.0, offset=0.0),
        RED: GainOffset(gain=1.0, offset=0.0),
        NIR: GainOffset(gain=1.0, offset=0.0),
        NARROW_NIR: GainOffset(gain=1.0, offset=0.0),
    },  # 2017v0
    Constellation.QB: {
        PAN: GainOffset(gain=0.870, offset=-1.491),
        BLUE: GainOffset(gain=1.105, offset=-2.820),
        GREEN: GainOffset(gain=1.071, offset=-3.338),
        RED: GainOffset(gain=1.060, offset=-2.954),
        NIR: GainOffset(gain=1.020, offset=-4.722),
        NARROW_NIR: GainOffset(gain=1.020, offset=-4.722),
    },  # 2016v0.Int
    Constellation.WV01: {
        PAN: GainOffset(gain=1.016, offset=-1.824),
    },  # 2016v0.Int
}
"""
The TDI specific abscalfactor and effectiveBandwidth are delivered with the imagery in the metadata file. The
digital number, DN, is the pixel value found in the imagery. The Gain and Offset are the absolute radiometric
calibration band dependent adjustment factors that are given in Table 1. Note that these are not necessarily
stagnant values and they are revisited annually.

Only using last calibration as the Maxar sensors have been found to be very stable throughout their lifetime.

See `here <https://apollomapping.com/image_downloads/Maxar_AbsRadCalDataSheet2018v0.pdf>`_ for the values.
"""

_MAXAR_E0 = {
    Constellation.GE01: {
        PAN: 1610.73,
        BLUE: 1993.18,
        GREEN: 1828.83,
        RED: 1491.49,
        NIR: 1022.58,
        NARROW_NIR: 1022.58,
    },
    Constellation.WV02: {
        PAN: 1571.36,
        CA: 1773.81,
        BLUE: 2007.27,
        GREEN: 1829.62,
        YELLOW: 1701.85,
        RED: 1538.85,
        VRE_1: 1346.09,
        VRE_2: 1346.09,
        VRE_3: 1346.09,
        NIR: 1053.21,
        NARROW_NIR: 1053.21,
        WV: 856.599,
    },
    Constellation.WV03: {
        PAN: 1574.41,
        CA: 1757.89,
        BLUE: 2004.61,
        GREEN: 1830.18,
        YELLOW: 1712.07,
        RED: 1535.33,
        VRE_1: 1348.08,
        VRE_2: 1348.08,
        VRE_3: 1348.08,
        NIR: 1055.94,
        NARROW_NIR: 1055.94,
        WV: 858.77,
    },
    Constellation.WV04: {
        PAN: 1608.01,
        BLUE: 2009.45,
        GREEN: 1831.88,
        RED: 1492.12,
        NIR: 937.8,
        NARROW_NIR: 937.8,
    },
    Constellation.QB: {
        PAN: 1370.92,
        BLUE: 1949.59,
        GREEN: 1823.64,
        RED: 1553.78,
        NIR: 1102.85,
        NARROW_NIR: 1102.85,
    },
    Constellation.WV01: {
        PAN: 1478.62,
    },
}
"""
Esun is the band-averaged Solar exoatmospheric irradiance @1AU (see Slide 7&8). DG calibration team uses the Thuillier 2003 solar curve for their calculations.
See `here <https://apollomapping.com/image_downloads/Maxar_AbsRadCalDataSheet2018v0.pdf>`_ for the values.
"""


@unique
class MaxarProductType(ListEnum):
    """
    Maxar product types.

    See `here <https://dg-cms-uploads-production.s3.amazonaws.com/uploads/document/file/106/ISD_External.pdf>`_ (p. 26)
    """

    Basic = "System Ready"
    """
    Basic Imagery, also known as System-Ready data.
    Corresponds to a Level 1B.

    Not available in EOReader.
    """

    Standard = "View Ready"
    """
    Standard Imagery, also known as View-Ready data (previously Ortho-Ready Standard).
    Corresponds to a Level 2A (Standard2A or ORStandard2A)
    """

    Ortho = "Map Ready"
    """
    Orthorectified Standard Imagery, also known as Map-Ready data (previously Standard Imagery).
    Corresponds to a Level 3

    NMAS mapping scale of the Orthorectified Product:

    - Level 3A: “1:50,000”
    - Level 3D: “1:12,000”
    - Level 3E: “1:10,000”
    - Level 3F: “1:5,000”
    - Level 3G: “1:4,800”
    - Level 3X: “Custom”
    """

    DEM = "DEM"
    """
    DEM product type.

    Not available in EOReader.
    """

    Stereo = "Stereo"
    """
    Stereo product type.

    Not available in EOReader.
    """


@unique
class MaxarBandId(ListEnum):
    """
    Maxar products band ID

    See `here <https://dg-cms-uploads-production.s3.amazonaws.com/uploads/document/file/106/ISD_External.pdf>`_ (p. 24)
    """

    P = "Panchromatic"
    """
    Panchromatic
    """

    Multi = "Multi Spectral"
    """
    All VNIR Multi-spectral bands (4 for QB02,GE01 and 8 for WV02, WV03)
    """

    N = "NIR"
    """
    Near-InfraRed
    """

    R = "RED"
    """
    Red
    """

    G = "GREEN"
    """
    Green
    """

    B = "BLUE"
    """
    Blue
    """

    RGB = "RGB"
    """
    Red + Green + Blue
    Pan-sharpened color images, stored at the panchromatic spatial resolution.
    """

    NRG = "NRG"
    """
    Near-IR + Red + Green
    Pan-sharpened color images, stored at the panchromatic spatial resolution.
    """

    BGRN = "BGRN Pan-Sharpened"
    """
    Blue + Green + Red + Near-IR
    Pan-sharpened color images, stored at the panchromatic spatial resolution.
    """

    # Only for WorldView-2 and WorldView-3
    N2 = "NIR2"
    """
    NIR2
    """

    RE = "Red-Edge"
    """
    NIR2
    """

    Y = "Yellow"
    """
    Yellow
    """

    C = "Coastal"
    """
    Coastal
    """

    MS1 = "Multi Spectral 1"
    """
    First 4 bands (N,R,G,B)
    """

    MS2 = "Multi Spectral 2"
    """
    Second 4 bands (N2,RE,Y,C)
    """


@unique
class MaxarSatId(ListEnum):
    """
    Maxar products satellite IDs

    See `here <https://dg-cms-uploads-production.s3.amazonaws.com/uploads/document/file/106/ISD_External.pdf>`_ (p. 29)
    """

    QB02 = "Quickbird"
    """
    Quickbird
    """

    GE01 = "GeoEye-1"
    """
    GeoEye-1
    """

    WV01 = "WorldView-1"
    """
    WorldView-1
    """

    WV02 = "WorldView-2"
    """
    WorldView-2
    """

    WV03 = "WorldView-3"
    """
    WorldView-3
    """

    WV04 = "WorldView-4"
    """
    WorldView-4
    """


class MaxarProduct(VhrProduct):
    """
    Super Class of Maxar products.
    See `here <https://earth.esa.int/eogateway/documents/20142/37627/DigitalGlobe-Standard-Imagery.pdf>`_
    for more information.
    """

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._raw_units = RawUnits.DN
        self._has_cloud_cover = True
        self.needs_extraction = False
        self._proj_prod_type = [MaxarProductType.Standard]

        # Not exact resolutions but usual ones
        self._pan_res = 0.5
        self._ms_res = 2.0

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _set_instrument(self) -> None:
        """
        Set instrument

        WV01: https://earth.esa.int/eogateway/missions/worldview-1
        WV02: https://earth.esa.int/eogateway/missions/worldview-2
        WV03: https://earth.esa.int/eogateway/missions/worldview-3
        WV04: https://space.oscar.wmo.int/satellites/view/worldview_4
        Quickbird: https://earth.esa.int/eogateway/missions/quickbird-2
        GeoEye: https://earth.esa.int/eogateway/missions/worldview-3
        """
        if self.constellation == Constellation.WV01:
            # WorldView-60 camera (WV60)
            self.instrument = "WV60"
        elif self.constellation in [Constellation.WV02, Constellation.WV03]:
            # WorldView-110 camera (WV110)
            self.instrument = "WV110"
        elif self.constellation == Constellation.WV04:
            # SpaceView-110 camera
            self.instrument = "SpaceView-110 camera"
        elif self.constellation == Constellation.QB:
            # Ball Global Imaging System 2000
            self.instrument = "BGIS-2000"
        elif self.constellation == Constellation.GE01:
            # GeoEye Imaging System (GIS)
            self.instrument = "GIS"

    def _get_constellation(self) -> Constellation:
        """Getter of the constellation"""
        # Maxar products are all similar, we must check into the metadata to know the constellation
        root, _ = self.read_mtd()
        constellation_id = root.findtext(".//IMAGE/SATID")
        if not constellation_id:
            raise InvalidProductError("Cannot find SATID in the metadata file")
        constellation_id = getattr(MaxarSatId, constellation_id).name
        return getattr(Constellation, constellation_id)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        # Band combination
        root, _ = self.read_mtd()
        band_combi = root.findtext(".//IMD/BANDID")
        if not band_combi:
            raise InvalidProductError("Cannot find from BANDID in the metadata file")
        self.band_combi = getattr(MaxarBandId, band_combi)

        # Post init done by the super class
        super()._post_init(**kwargs)

    @abstractmethod
    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """

        # Band combination
        root, _ = self.read_mtd()
        gsd = root.findtext(".//MAP_PROJECTED_PRODUCT/PRODUCTGSD")
        if not gsd:
            raise InvalidProductError(
                "Cannot find PRODUCTGSD type in the metadata file"
            )
        self.pixel_size = float(gsd)

    def _get_spectral_bands(self) -> dict:
        """
        https://docs.sentinel-hub.com/api/latest/static/files/data/maxar/world-view/resources/brochures/EUSI_Satellite_Booklet_digital.pdf
        https://resources.maxar.com/optical-imagery/multispectral-reference-guide

        Returns:
            dict: Maxar spectral bands
        """
        # Create spectral bands
        if self.constellation in [Constellation.WV02, Constellation.WV03]:
            spectral_bands = {
                "pan": SpectralBand(
                    eoreader_name=PAN,
                    **{
                        NAME: "PAN",
                        ID: 1,
                        GSD: self._pan_res,
                        WV_MIN: 450,
                        WV_MAX: 800,
                    },
                ),
                "ca": SpectralBand(
                    eoreader_name=CA,
                    **{
                        NAME: "COASTAL BLUE",
                        ID: 1,
                        GSD: self._ms_res,
                        WV_MIN: 400,
                        WV_MAX: 450,
                    },
                ),
                "blue": SpectralBand(
                    eoreader_name=BLUE,
                    **{
                        NAME: "BLUE",
                        ID: 2,
                        GSD: self._ms_res,
                        WV_MIN: 450,
                        WV_MAX: 510,
                    },
                ),
                "green": SpectralBand(
                    eoreader_name=GREEN,
                    **{
                        NAME: "GREEN",
                        ID: 3,
                        GSD: self._ms_res,
                        WV_MIN: 510,
                        WV_MAX: 580,
                    },
                ),
                "yellow": SpectralBand(
                    eoreader_name=YELLOW,
                    **{
                        NAME: "YELLOW",
                        ID: 7,
                        GSD: self._ms_res,
                        WV_MIN: 585,
                        WV_MAX: 625,
                    },
                ),
                "red": SpectralBand(
                    eoreader_name=RED,
                    **{NAME: "RED", ID: 5, GSD: self._ms_res, WV_MIN: 630, WV_MAX: 690},
                ),
                "vre": SpectralBand(
                    eoreader_name=VRE_1,
                    **{
                        NAME: "RED EDGE",
                        ID: 6,
                        GSD: self._ms_res,
                        WV_MIN: 705,
                        WV_MAX: 745,
                    },
                ),
                "nir": SpectralBand(
                    eoreader_name=NIR,
                    **{
                        NAME: "NIR1",
                        ID: 7,
                        GSD: self._ms_res,
                        WV_MIN: 770,
                        WV_MAX: 895,
                    },
                ),
                "wv": SpectralBand(
                    eoreader_name=WV,
                    **{
                        NAME: "NIR2",
                        ID: 8,
                        GSD: self._ms_res,
                        WV_MIN: 860,
                        WV_MAX: 1040,
                    },
                ),
            }
        elif self.constellation == Constellation.QB:
            spectral_bands = {
                "pan": SpectralBand(
                    eoreader_name=PAN,
                    **{
                        NAME: "PAN",
                        ID: 1,
                        GSD: self._pan_res,
                        WV_MIN: 405,
                        WV_MAX: 1053,
                    },
                ),
                "blue": SpectralBand(
                    eoreader_name=BLUE,
                    **{
                        NAME: "BLUE",
                        ID: 1,
                        GSD: self._ms_res,
                        WV_MIN: 430,
                        WV_MAX: 545,
                    },
                ),
                "green": SpectralBand(
                    eoreader_name=GREEN,
                    **{
                        NAME: "GREEN",
                        ID: 2,
                        GSD: self._ms_res,
                        WV_MIN: 466,
                        WV_MAX: 620,
                    },
                ),
                "red": SpectralBand(
                    eoreader_name=RED,
                    **{NAME: "RED", ID: 3, GSD: self._ms_res, WV_MIN: 590, WV_MAX: 710},
                ),
                "nir": SpectralBand(
                    eoreader_name=NIR,
                    **{
                        NAME: "NIR1",
                        ID: 4,
                        GSD: self._ms_res,
                        WV_MIN: 715,
                        WV_MAX: 918,
                    },
                ),
            }
        elif self.constellation == Constellation.WV01:
            spectral_bands = {
                "pan": SpectralBand(
                    eoreader_name=PAN,
                    **{
                        NAME: "PAN",
                        ID: 1,
                        GSD: self._pan_res,
                        WV_MIN: 400,
                        WV_MAX: 900,
                    },
                )
            }
        elif self.constellation in [Constellation.GE01, Constellation.WV04]:
            spectral_bands = {
                "pan": SpectralBand(
                    eoreader_name=PAN,
                    **{
                        NAME: "PAN",
                        ID: 1,
                        GSD: self._pan_res,
                        WV_MIN: 450,
                        WV_MAX: 800,
                    },
                ),
                "blue": SpectralBand(
                    eoreader_name=BLUE,
                    **{
                        NAME: "BLUE",
                        ID: 1 if self.constellation == Constellation.GE01 else 3,
                        GSD: self._ms_res,
                        WV_MIN: 450,
                        WV_MAX: 510,
                    },
                ),
                "green": SpectralBand(
                    eoreader_name=GREEN,
                    **{
                        NAME: "GREEN",
                        ID: 2,
                        GSD: self._ms_res,
                        WV_MIN: 510,
                        WV_MAX: 580,
                    },
                ),
                "red": SpectralBand(
                    eoreader_name=RED,
                    **{
                        NAME: "RED",
                        ID: 3 if self.constellation == Constellation.GE01 else 1,
                        GSD: self._ms_res,
                        WV_MIN: 655,
                        WV_MAX: 690,
                    },
                ),
                "nir": SpectralBand(
                    eoreader_name=NIR,
                    **{
                        NAME: "NIR1",
                        ID: 4,
                        GSD: self._ms_res,
                        WV_MIN: 780,
                        WV_MAX: 920,
                    },
                ),
            }
        else:
            raise InvalidProductError(f"Unknown platform: {self.constellation}")

        return spectral_bands

    def _get_band_map(self, **kwargs) -> dict:
        """Get band map"""
        # Open spectral bands
        pan = kwargs.get("pan")
        blue = kwargs.get("blue")
        green = kwargs.get("green")
        red = kwargs.get("red")
        nir = kwargs.get("nir")
        ca = kwargs.get("ca")
        vre = kwargs.get("vre")
        yellow = kwargs.get("yellow")
        wv = kwargs.get("wv")

        # Manage bands of the product
        if self.band_combi in [
            MaxarBandId.P,
            MaxarBandId.N,
            MaxarBandId.R,
            MaxarBandId.G,
            MaxarBandId.B,
            MaxarBandId.N2,
            MaxarBandId.RE,
            MaxarBandId.Y,
            MaxarBandId.C,
        ]:
            band_map = {PAN: pan.update(id=1, gsd=self.pixel_size)}
        elif self.band_combi == MaxarBandId.RGB:
            band_map = {
                RED: red.update(id=1, gsd=self.pixel_size),
                GREEN: green.update(id=2, gsd=self.pixel_size),
                BLUE: blue.update(id=3, gsd=self.pixel_size),
            }
        elif self.band_combi == MaxarBandId.NRG:
            band_map = {
                NIR: nir.update(id=1, gsd=self.pixel_size),
                NARROW_NIR: nir.update(id=1, gsd=self.pixel_size),
                RED: red.update(id=2, gsd=self.pixel_size),
                GREEN: green.update(id=3, gsd=self.pixel_size),
            }
        elif self.band_combi == MaxarBandId.BGRN:
            band_map = {
                BLUE: blue.update(id=1, gsd=self.pixel_size),
                GREEN: green.update(id=2, gsd=self.pixel_size),
                RED: red.update(id=3, gsd=self.pixel_size),
                NIR: nir.update(id=4, gsd=self.pixel_size),
                NARROW_NIR: nir.update(id=4, gsd=self.pixel_size),
            }
        elif self.band_combi == MaxarBandId.MS1:
            band_map = {
                NIR: nir.update(id=1, gsd=self.pixel_size),
                NARROW_NIR: nir.update(id=1, gsd=self.pixel_size),
                RED: red.update(id=2, gsd=self.pixel_size),
                GREEN: green.update(id=3, gsd=self.pixel_size),
                BLUE: blue.update(id=4, gsd=self.pixel_size),
            }
        elif self.band_combi == MaxarBandId.MS2:
            band_map = {
                WV: wv.update(id=1, gsd=self.pixel_size),
                VRE_1: vre.update(id=2, gsd=self.pixel_size),
                VRE_2: vre.update(id=2, gsd=self.pixel_size),
                VRE_3: vre.update(id=2, gsd=self.pixel_size),
                YELLOW: yellow.update(id=3, gsd=self.pixel_size),
                CA: ca.update(id=4, gsd=self.pixel_size),
            }
        elif self.band_combi == MaxarBandId.Multi:
            if self.constellation_id in (MaxarSatId.WV02.name, MaxarSatId.WV03.name):
                band_map = {
                    CA: ca.update(id=1, gsd=self.pixel_size),
                    BLUE: blue.update(id=2, gsd=self.pixel_size),
                    GREEN: green.update(id=3, gsd=self.pixel_size),
                    YELLOW: yellow.update(id=4, gsd=self.pixel_size),
                    RED: red.update(id=5, gsd=self.pixel_size),
                    VRE_1: vre.update(id=6, gsd=self.pixel_size),
                    VRE_2: vre.update(id=6, gsd=self.pixel_size),
                    VRE_3: vre.update(id=6, gsd=self.pixel_size),
                    NIR: nir.update(id=7, gsd=self.pixel_size),
                    NARROW_NIR: nir.update(id=7, gsd=self.pixel_size),
                    WV: wv.update(id=8, gsd=self.pixel_size),
                }
            else:
                band_map = {
                    NIR: nir.update(id=1, gsd=self.pixel_size),
                    NARROW_NIR: nir.update(id=1, gsd=self.pixel_size),
                    RED: red.update(id=2, gsd=self.pixel_size),
                    GREEN: green.update(id=3, gsd=self.pixel_size),
                    BLUE: blue.update(id=4, gsd=self.pixel_size),
                }
        else:
            raise InvalidProductError(
                f"Unusual band combination: {self.band_combi.name}"
            )

        return band_map

    def _map_bands(self):
        """
        Map bands
        """
        self.bands.map_bands(self._get_band_map(**self._get_spectral_bands()))

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, _ = self.read_mtd()
        prod_type = root.findtext(".//IMD/PRODUCTTYPE")
        if not prod_type:
            raise InvalidProductError(
                "Cannot find the PRODUCTTYPE in the metadata file"
            )
        self.product_type = getattr(MaxarProductType, prod_type)
        if self.product_type not in (MaxarProductType.Ortho, MaxarProductType.Standard):
            raise NotImplementedError(
                f"For now, "
                f"only {MaxarProductType.Ortho.value, MaxarProductType.Standard.value} "
                f"product types are supported for Maxar products."
            )

        if self.product_type == MaxarProductType.Standard:
            self.is_ortho = False

    def _get_raw_crs(self) -> riocrs.CRS:
        """
        Get raw CRS of the tile

        Returns:
            rasterio.crs.CRS: CRS object
        """
        # Open metadata
        root, _ = self.read_mtd()

        # Get CRS
        map_proj_name = root.findtext(".//MAPPROJNAME")
        if not map_proj_name:
            raise InvalidProductError("Cannot find MAPPROJNAME in the metadata file")
        if map_proj_name == "Geographic (Lat/Long)":
            crs = riocrs.CRS.from_string("EPSG:4326")
        elif map_proj_name == "UTM":
            map_hemi = root.findtext(".//MAPHEMI")
            map_zone = root.findtext(".//MAPZONE")
            if not map_hemi or not map_zone:
                raise InvalidProductError(
                    "Cannot find MAPHEMI or MAPZONE type in the metadata file"
                )
            crs = riocrs.CRS.from_string(
                f"EPSG:32{6 if map_hemi == 'N' else 7}{map_zone}"
            )
        else:
            raise NotImplementedError(
                "Only Geographic or UTM map projections are supported yet"
            )

        return crs

    @cache
    def crs(self) -> riocrs.CRS:
        """
        Get UTM projection of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.crs()
            CRS.from_epsg(32618)

        Returns:
            rasterio.crs.CRS: CRS object
        """
        # Get Raw CRS
        raw_crs = self._get_raw_crs()

        # Get CRS
        if raw_crs.is_geographic:
            # Open metadata
            root, _ = self.read_mtd()

            # Get the origin lon lat
            lon = float(root.findtext(".//ORIGINX"))
            lat = float(root.findtext(".//ORIGINY"))

            # Compute UTM crs from center long/lat
            utm = vectors.to_utm_crs(lon, lat)
        else:
            utm = raw_crs

        return utm

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get real footprint in UTM of the products (without nodata, in french == emprise utile)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
                                                         gml_id  ...                                           geometry
            0  source_image_footprint-DS_PHR1A_20200511023124...  ...  POLYGON ((707025.261 9688613.833, 707043.276 9...
            [1 rows x 3 columns]

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        # If ortho -> nodata is not set !
        if self.is_ortho:
            # Get footprint of the first band of the stack
            footprint_dezoom = 10
            arr = rasters.read(
                self.get_default_band_path(),
                resolution=self.pixel_size * footprint_dezoom,
                indexes=[1],
            )

            # Vectorize the nodata band
            footprint = rasters.vectorize(
                arr, values=0, keep_values=False, dissolve=True
            )
            footprint = geometry.get_wider_exterior(footprint.convex_hull)
        else:
            # If not ortho -> default band has been orthorectified and nodata will be set
            footprint = rasters.get_footprint(self.get_default_band_path())

        return footprint.to_crs(self.crs())

    @cache
    def extent(self, **kwargs) -> gpd.GeoDataFrame:
        """
        Get UTM extent of the tile.

        Returns:
            gpd.GeoDataFrame: Extent in UTM
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Compute extent corners
        default_extent = root.find(".//MAP_PROJECTED_PRODUCT")
        ul_corner = float(default_extent.findtext("ULX")), float(
            default_extent.findtext("ULY")
        )
        ur_corner = float(default_extent.findtext("URX")), float(
            default_extent.findtext("URY")
        )
        lr_corner = float(default_extent.findtext("LRX")), float(
            default_extent.findtext("LRY")
        )
        ll_corner = float(default_extent.findtext("LLX")), float(
            default_extent.findtext("LLY")
        )
        corners = [ul_corner, ur_corner, lr_corner, ll_corner]

        raw_extent = gpd.GeoDataFrame(
            geometry=[Polygon(corners)],
            crs=self._get_raw_crs(),
        )

        return raw_extent.to_crs(self.crs())

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format  :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 5, 11, 2, 31, 58)
            >>> prod.get_datetime(as_datetime=False)
            '20200511T023158'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()
            datetime_str = root.findtext(".//EARLIESTACQTIME")
            if not datetime_str:
                datetime_str = root.findtext(".//FIRSTLINETIME")
                if not datetime_str:
                    raise InvalidProductError(
                        "Cannot find EARLIESTACQTIME or FIRSTLINETIME in the metadata file."
                    )

            # Convert to datetime
            datetime_str = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%fZ")

            if not as_datetime:
                datetime_str = datetime_str.strftime(DATETIME_FMT)

        else:
            datetime_str = self.datetime
            if not as_datetime:
                datetime_str = datetime_str.strftime(DATETIME_FMT)

        return datetime_str

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        return path.get_filename(self._get_tile_path())

    @cache
    def get_mean_sun_angles(self) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1A_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (45.6624568841367, 30.219881316357643)

        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open zenith and azimuth angle
        try:
            elev_angle = float(root.findtext(".//MEANSUNEL"))
            azimuth_angle = float(root.findtext(".//MEANSUNAZ"))
        except TypeError:
            raise InvalidProductError("Azimuth or Zenith angles not found in metadata!")

        # From elevation to zenith
        zenith_angle = 90.0 - elev_angle

        return azimuth_angle, zenith_angle

    @cache
    def get_mean_viewing_angles(self) -> (float, float, float):
        """
        Get Mean Viewing angles (azimuth, off-nadir and incidence angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_viewing_angles()

        Returns:
            (float, float, float): Mean azimuth, off-nadir and incidence angles
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open zenith and azimuth angle
        try:
            az = float(root.findtext(".//MEANSATAZ"))
            incidence_angle = 90 - float(root.findtext(".//MEANSATEL"))
            off_nadir = float(root.findtext(".//MEANOFFNADIRVIEWANGLE"))
        except TypeError:
            raise InvalidProductError(
                "MEANSATAZ, MEANSATEL or MEANOFFNADIRVIEWANGLE angles not found in metadata!"
            )

        return az, off_nadir, incidence_angle

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces as a dict
        """
        mtd_from_path = f"{self.name}.XML"
        mtd_archived = rf"{self.name}\.XML"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        # Convert DN into radiance
        band_arr = self._dn_to_toa_rad(band_arr, band)

        # Convert radiance into reflectance
        band_arr = self._toa_rad_to_toa_refl(band_arr, band)

        # To float32
        if band_arr.dtype != np.float32:
            band_arr = band_arr.astype(np.float32)

        return band_arr

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        """
        return False

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        if bands:
            LOGGER.warning("Maxar products do not provide any cloud file")

        return {}

    def _get_tile_path(self) -> AnyPathType:
        """
        Get the DIMAP filepath

        Returns:
            AnyPathType: DIMAP filepath
        """
        return self._get_path(extension="TIL")

    def _dn_to_toa_rad(self, dn_arr: xr.DataArray, band: BandNames) -> xr.DataArray:
        """
        Compute DN to TOA radiance

        See
        `here <https://apollomapping.com/image_downloads/Maxar_AbsRadCalDataSheet2018v0.pdf>`_
        for more information.

        Args:
            dn_arr (xr.DataArray): DN array
            band (BandNames): Band

        Returns:
            xr.DataArray: TOA Radiance array
        """
        band_mtd_str = f"BAND_{_MAXAR_BAND_MTD[band]}"

        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open absolute calibration factor and the effective bandwidth
        try:
            band_mtd = root.find(f".//{band_mtd_str}")
            abs_factor = float(band_mtd.findtext(".//ABSCALFACTOR"))
            effective_bandwidth = float(band_mtd.findtext(".//EFFECTIVEBANDWIDTH"))
        except TypeError:
            raise InvalidProductError(
                "ABSCALFACTOR or EFFECTIVEBANDWIDTH not found in metadata!"
            )

        # Get constellation-specific gain and offset (latest)
        gain, offset = _MAXAR_GAIN_OFFSET[self.constellation][band]

        # Compute the coefficient converting DN in TOA radiance
        coeff = gain * abs_factor / effective_bandwidth

        # LOGGER.debug(f"DN to rad coeff = {coeff}")
        return dn_arr.copy(data=coeff * dn_arr.data + offset)

    def _toa_rad_to_toa_refl(
        self, rad_arr: xr.DataArray, band: BandNames
    ) -> xr.DataArray:
        """
        Compute TOA reflectance from TOA radiance

        See
        `here <https://apollomapping.com/image_downloads/Maxar_AbsRadCalDataSheet2018v0.pdf>`_
        for more information.

        WARNING: in this formula, d**2 = 1 / sqrt(dt) !

        Args:
            rad_arr (xr.DataArray): TOA Radiance array
            band (BandNames): Band

        Returns:
            xr.DataArray: TOA Reflectance array
        """

        # Compute the coefficient converting TOA radiance in TOA reflectance
        dt = self._sun_earth_distance_variation() ** 2
        _, sun_zen = self.get_mean_sun_angles()
        rad_sun_zen = np.deg2rad(sun_zen)
        e0 = _MAXAR_E0[self.constellation][band]
        toa_refl_coeff = np.pi / (e0 * dt * np.cos(rad_sun_zen))

        # LOGGER.debug(f"rad to refl coeff = {toa_refl_coeff}")
        return rad_arr.copy(data=toa_refl_coeff * rad_arr)

    @cache
    def get_cloud_cover(self) -> float:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_cloud_cover()
            55.5

        Returns:
            float: Cloud cover as given in the metadata
        """

        # Get MTD XML file
        root, _ = self.read_mtd()

        # Get the cloud cover
        try:
            cc = float(root.findtext(".//CLOUDCOVER"))

        except TypeError:
            raise InvalidProductError("CLOUDCOVER not found in metadata!")

        # Manage the case with cloud_cover == -999.0
        # i.e. 17APR05171409-M1BS_R1C1-000000010003_01_P001 (WV04)
        if cc < 0.0:
            cc = None

        return cc

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = self.path / path.get_archived_path(
                    self.path, file_regex=r".*BROWSE\.JPG"
                )
            else:
                quicklook_path = next(self.path.glob("*BROWSE.JPG"))
            quicklook_path = str(quicklook_path)
        except (StopIteration, FileNotFoundError):
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    def _get_job_id(self) -> str:
        """
        Get VHR job ID

        Returns:
            str: VHR product ID
        """
        return self.split_name[0].split("-")[-1]
